from __future__ import absolute_import

import logging
import socket
import uuid

from amqp.exceptions import NotFound
from kombu.common import maybe_declare
from kombu.messaging import Queue

from nameko import config, serialization
from nameko.amqp.consume import Consumer
from nameko.amqp.publish import Publisher, get_connection
from nameko.constants import (
    AMQP_SSL_CONFIG_KEY, AMQP_URI_CONFIG_KEY, CALL_ID_STACK_CONTEXT_KEY,
    DEFAULT_AMQP_URI, DEFAULT_HEARTBEAT, DEFAULT_PREFETCH_COUNT,
    HEARTBEAT_CONFIG_KEY, PREFETCH_COUNT_CONFIG_KEY
)
from nameko.containers import new_call_id
from nameko.exceptions import ReplyQueueExpiredWithPendingReplies, RpcTimeout
from nameko.messaging import encode_to_headers
from nameko.rpc import (
    RESTRICTED_PUBLISHER_OPTIONS, RPC_REPLY_QUEUE_TEMPLATE,
    RPC_REPLY_QUEUE_TTL, Client, get_rpc_exchange
)


_logger = logging.getLogger(__name__)


class ReplyListener(object):
    """ Single-threaded listener for RPC replies.

    Creates a queue and consumes from it on demand. RPC requests
    should register their `correlation_id` with
    :meth:`~ReplyListener.register_for_reply` in order for the `ReplyListener`
    to capture the reply.
    """
    class ReplyConsumer(Consumer):
        """ Subclass Consumer to add disconnection check
        """

        def __init__(self, check_for_lost_replies, *args, **kwargs):
            self.check_for_lost_replies = check_for_lost_replies
            super(ReplyListener.ReplyConsumer, self).__init__(*args, **kwargs)

        def get_consumers(self, consumer_cls, channel):
            """
            Check for messages lost while the reply listener was disconnected
            from the broker.
            """
            self.check_for_lost_replies()

            return super(ReplyListener.ReplyConsumer, self).get_consumers(
                consumer_cls, channel
            )

    consumer_cls = ReplyConsumer

    def __init__(self, queue, timeout=None, uri=None, ssl=None, **kwargs):
        self.queue = queue
        self.timeout = timeout

        self.amqp_uri = uri or config.get(AMQP_URI_CONFIG_KEY, DEFAULT_AMQP_URI)
        self.ssl = ssl if ssl is not None else config.get(AMQP_SSL_CONFIG_KEY)

        self.pending = {}

    def start(self):
        heartbeat = config.get(HEARTBEAT_CONFIG_KEY, DEFAULT_HEARTBEAT)
        prefetch_count = config.get(
            PREFETCH_COUNT_CONFIG_KEY, DEFAULT_PREFETCH_COUNT
        )
        accept = serialization.setup().accept

        queues = [self.queue]
        callbacks = [self.handle_message]

        self.consumer = self.consumer_cls(
            self.check_for_lost_replies, self.amqp_uri, ssl=self.ssl,
            queues=queues, callbacks=callbacks,
            heartbeat=heartbeat, prefetch_count=prefetch_count, accept=accept
        )

        # must declare queue because the consumer doesn't start right away
        with self.consumer.connection as conn:
            maybe_declare(self.queue, conn.channel())

    def stop(self):
        self.consumer.stop()

    def check_for_lost_replies(self):
        if self.pending:
            try:
                with get_connection(self.amqp_uri, ssl=self.ssl) as conn:
                    self.queue.bind(conn).queue_declare(passive=True)
            except NotFound:
                raise ReplyQueueExpiredWithPendingReplies(
                    "Lost replies for correlation ids:\n{}".format(
                        "\n".join(self.pending.keys())
                    )
                )

    def register_for_reply(self, correlation_id):
        """ Register an RPC call with the given `correlation_id` for a reply.

        Returns a function that can be used to retrieve the reply, blocking
        until it has been received.
        """
        self.pending[correlation_id] = None
        return lambda: self.consume_reply(correlation_id)

    def consume_reply(self, correlation_id):
        """ Consume from the reply queue until the reply for the given
        `correlation_id` is received.
        """
        # return error if correlation_id not pending? (new feature)
        if self.consumer.should_stop:
            raise RuntimeError("Stopped and can no longer be used")

        while not self.pending.get(correlation_id):
            try:
                next(self.consumer.consume(timeout=self.timeout))
            except socket.timeout:
                raise RpcTimeout()
        return self.pending.pop(correlation_id)

    def handle_message(self, body, message):
        self.consumer.ack_message(message)

        correlation_id = message.properties.get('correlation_id')
        if correlation_id not in self.pending:
            _logger.debug("Unknown correlation id: %s", correlation_id)
            return

        self.pending[correlation_id] = body


class ClusterRpcClient(object):
    """
    Single-threaded RPC client to a cluster of services. The target service
    and method are specified with attributes.

    Method calls on the local object are converted into RPC calls to the
    target service.

    Enables services not hosted by nameko to make RPC requests to a nameko
    cluster. It is commonly used as a context manager but may also be manually
    started and stopped.

    *Usage*

    As a context manager::

        with ClusterRpc() as client:
            client.target_service.method()
            client.other_service.method()

    The equivalent call, manually starting and stopping::

        client = ClusterRpc()
        client = client.start()
        try:
            client.target_service.method()
            client.other_service.method()
        finally:
            client.stop()

    If you call ``start()`` you must eventually call ``stop()`` to close the
    connection to the broker.

    You may also supply ``context_data``, a dictionary of data to be
    serialised into the AMQP message headers.

    When the name of the service is not legal in Python, you can also
    use a dict-like syntax::

        with ClusterRpc() as client:
            client['service-name'].method()
            client['other-service'].method()

    """

    publisher_cls = Publisher

    def __init__(
        self, context_data=None, timeout=None, **publisher_options
    ):
        self.uuid = str(uuid.uuid4())

        exchange = get_rpc_exchange()

        queue_name = RPC_REPLY_QUEUE_TEMPLATE.format(
            "standalone_rpc_client", self.uuid
        )
        queue = Queue(
            queue_name,
            exchange=exchange,
            routing_key=self.uuid,
            queue_arguments={
                'x-expires': RPC_REPLY_QUEUE_TTL
            }
        )

        self.amqp_uri = publisher_options.pop(
            'uri', config.get(AMQP_URI_CONFIG_KEY, DEFAULT_AMQP_URI)
        )
        self.ssl = publisher_options.pop(
            'ssl', config.get(AMQP_SSL_CONFIG_KEY)
        )

        self.reply_listener = ReplyListener(queue, timeout=timeout)

        serialization_config = serialization.setup()
        self.serializer = publisher_options.pop(
            'serializer', serialization_config.serializer
        )

        for option in RESTRICTED_PUBLISHER_OPTIONS:
            publisher_options.pop(option, None)

        publisher = self.publisher_cls(
            self.amqp_uri,
            ssl=self.ssl,
            exchange=exchange,
            serializer=self.serializer,
            declare=[self.reply_listener.queue],
            reply_to=self.reply_listener.queue.routing_key,
            **publisher_options
        )

        context_data = context_data or {}

        def publish(*args, **kwargs):

            context_data[CALL_ID_STACK_CONTEXT_KEY] = [
                'standalone_rpc_client.{}.{}'.format(self.uuid, new_call_id())
            ]

            extra_headers = kwargs.pop('extra_headers')
            extra_headers.update(encode_to_headers(context_data))

            publisher.publish(
                *args, extra_headers=extra_headers, **kwargs
            )

        get_reply = self.reply_listener.register_for_reply

        self.client = Client(publish, get_reply, context_data)

    def __enter__(self):
        return self.start()

    def __exit__(self, tpe, value, traceback):
        self.stop()

    def start(self):
        self.reply_listener.start()
        return self.client

    def stop(self):
        self.reply_listener.stop()


class ServiceRpcClient(ClusterRpcClient):
    """
    Single-threaded RPC client to a named service.

    As per :class:`~nameko.standalone.rpc.ClusterRpc` but with a pre-specified
    target service.
    """

    def __init__(self, service_name, *args, **kwargs):
        super(ServiceRpcClient, self).__init__(*args, **kwargs)
        self.client = getattr(self.client, service_name)


ClusterRpcProxy = ClusterRpcClient  # backwards compat
ServiceRpcProxy = ServiceRpcClient  # backwards compat



"""Custom click option decorators

Allow reuse of options across subcommands.

Reuse in other click based commands is also possible.

See nameko/cli/__init__.py for real use.
"""
from functools import partial

import click

from .click_paramtypes import HOST_PORT, KEY_VAL


option_broker = partial(
    click.option,
    "--broker",
    help="Deprecated option for setting up RabbitMQ broker URI.\n"
    "Use --define or --config and set AMQP_URI instead.",
    metavar="BROKER",
)
option_config_file = partial(
    click.option,
    "-c",
    "--config",
    "config_file",
    help="The YAML configuration file",
    type=click.File("rb"),
    metavar="CONFIG",
)
option_define = partial(
    click.option,
    "-d",
    "--define",
    help="Set config entry. Overrides value loaded from config file."
    " Can be used multiple times. Example: --define"
    " AMQP_URI=pyamqp://guest:guest@localhost",
    type=KEY_VAL,
    multiple=True,
    metavar="KEY=VALUE",
    callback=lambda ctx, param, value: dict(value),
)
option_backdoor_port = partial(
    click.option,
    "--backdoor-port",
    help="Specify a port number to host a backdoor, which can be connected to"
    " for an interactive interpreter within"
    " the running service process using `nameko backdoor`.",
    type=HOST_PORT,
)
option_interface = partial(
    click.option, "--interface", type=click.Choice(["bpython", "ipython", "plain"])
)



from __future__ import absolute_import

import json
import uuid
from collections import defaultdict

from eventlet.event import Event
from eventlet.queue import Queue

from nameko.exceptions import deserialize


def make_virtual_socket(host, port, path='/ws'):
    from websocket import WebSocketApp

    result_handlers = {}

    class Socket(object):

        def __init__(self):
            self._event_queues = defaultdict(Queue)

        def get_event_queue(self, event_type):
            return self._event_queues[event_type]

        def wait_for_event(self, event_type):
            return self.get_event_queue(event_type).get()

        def rpc(self, _method, **data):
            id = str(uuid.uuid4())
            event = Event()
            result_handlers[id] = event.send
            ws_app.send(json.dumps({
                'method': _method,
                'data': data,
                'correlation_id': id,
            }))

            rv = event.wait()
            if rv['success']:
                return rv['data']
            raise deserialize(rv['error'])

    sock = Socket()

    def on_message(ws, message):
        msg = json.loads(message)
        if msg['type'] == 'event':
            sock.get_event_queue(msg['event']).put((msg['event'], msg['data']))
        elif msg['type'] == 'result':
            result_id = msg['correlation_id']
            handler = result_handlers.pop(result_id, None)
            if handler is not None:
                handler(msg)

    ready_event = Event()

    def on_open(ws):
        ready_event.send(None)

    def on_error(ws, err):
        ready_event.send(err)

    ws_app = WebSocketApp(
        'ws://%s:%d%s' % (host, port, path),
        on_message=on_message,
        on_open=on_open,
        on_error=on_error,
    )

    def connect_socket():
        err = ready_event.wait()
        if err is not None:
            raise err  # (www.logilab.org/ticket/3207) pylint: disable=E0702
        return sock

    return ws_app, connect_socket



from nameko.events import EventDispatcher, event_handler
from nameko.rpc import rpc

class ServiceA:
    """ Event dispatching service. """
    name = "service_a"

    dispatch = EventDispatcher()

    @rpc
    def dispatching_method(self, payload):
        self.dispatch("event_type", payload)


class ServiceB:
    """ Event listening service. """
    name = "service_b"

    @event_handler("service_a", "event_type")
    def handle_event(self, payload):
        print("service b received:", payload)



import functools
import itertools
from time import sleep

import wrapt


class RetryDelay(object):
    def __init__(self, delay, backoff, max_delay):
        self.delay = delay
        self.backoff = backoff
        self.max_delay = max_delay

    def next(self):
        if self.backoff:
            self.delay *= self.backoff

        if self.max_delay:
            return min(self.delay, self.max_delay)

        return self.delay


def retry(
    wrapped=None, for_exceptions=Exception, max_attempts=3,
    delay=1, backoff=1, max_delay=None
):
    if wrapped is None:
        return functools.partial(
            retry, for_exceptions=for_exceptions, max_attempts=max_attempts,
            delay=delay, backoff=backoff, max_delay=max_delay
        )

    if max_attempts is None:
        max_attempts = float('inf')

    retry_delay = RetryDelay(delay, backoff, max_delay)

    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):

        counter = itertools.count()

        while True:
            try:
                return wrapped(*args, **kwargs)
            except for_exceptions:
                if next(counter) == max_attempts:
                    raise
                sleep(retry_delay.next())

    return wrapper(wrapped)  # pylint: disable=E1120



"""Custom click parameter types used in nameko.
"""
import click
import six
import yaml


class KeyValParamType(click.ParamType):
    """KEY=value formatted string. If only KEY is present, it gets value True.
    """

    name = "key_val"

    def convert(self, value, param, ctx):
        if not isinstance(value, six.string_types):
            self.fail("Value must by a string")
        if "=" in value:
            key, val = value.split("=", 1)
            val = yaml.safe_load(val)
            return key.strip(), val
        else:
            return value.strip(), True


KEY_VAL = KeyValParamType()


class HostPortParamType(click.ParamType):
    """[host]:port with default 'localhost' host. Port must be int.
    """

    name = "host_port"

    def convert(self, value, param, ctx):
        host = "localhost"
        port_str = ""
        if ":" in value:
            host, port_str = value.rsplit(":", 1)
        else:
            port_str = value
        try:
            port = int(port_str)
            return host, port
        except ValueError:
            self.fail("%s is not a valid port number" % port_str, param, ctx)


HOST_PORT = HostPortParamType()



import sys

import pytest
from mock import patch


@pytest.fixture
def command():
    from nameko.cli import cli

    def _command(*argv):
        with patch.object(sys, "argv", list(argv)):
            cli(standalone_mode=False)

    return _command



import pytest
from mock import Mock, call

from nameko import config
from nameko.constants import PARENT_CALLS_CONFIG_KEY
from nameko.containers import WorkerContext
from nameko.events import EventDispatcher, event_handler
from nameko.extensions import DependencyProvider
from nameko.rpc import ServiceRpc, rpc
from nameko.testing.services import entrypoint_hook, entrypoint_waiter
from nameko.testing.utils import DummyProvider, get_container


@pytest.mark.usefixtures("predictable_call_ids")
def test_worker_context_gets_stack(container_factory):

    class FooService(object):
        name = 'baz'

    container = container_factory(FooService)
    service = FooService()

    context = WorkerContext(container, service, DummyProvider("bar"))
    assert context.call_id == 'baz.bar.0'
    assert context.call_id_stack == ['baz.bar.0']

    # Build stack
    context = WorkerContext(
        container, service, DummyProvider("foo"),
        data={'call_id_stack': context.call_id_stack}
    )
    assert context.call_id == 'baz.foo.1'
    assert context.call_id_stack == ['baz.bar.0', 'baz.foo.1']

    # Long stack
    many_ids = [str(i) for i in range(10)]
    context = WorkerContext(
        container, service, DummyProvider("long"),
        data={'call_id_stack': many_ids}
    )
    expected = many_ids + ['baz.long.2']
    assert context.call_id_stack == expected


@pytest.mark.usefixtures("predictable_call_ids")
@config.patch({PARENT_CALLS_CONFIG_KEY: 1}, clear=True)
def test_short_call_stack(container_factory):

    class FooService(object):
        name = 'baz'

    container = container_factory(FooService)
    service = FooService()

    many_ids = [str(i) for i in range(100)]

    context = WorkerContext(
        container, service, DummyProvider("long"),
        data={'call_id_stack': many_ids}
    )
    # Trimmed stack
    assert context.call_id_stack == ['99', 'baz.long.0']


@pytest.fixture
def tracker():
    return Mock()


@pytest.fixture
def stack_logger(tracker):

    class StackLogger(DependencyProvider):

        def worker_setup(self, worker_ctx):
            tracker(worker_ctx.call_id_stack)

    return StackLogger


def test_call_id_stack(
    rabbit_config, predictable_call_ids, runner_factory, stack_logger, tracker
):
    StackLogger = stack_logger

    class Child(object):
        name = 'child'

        stack_logger = StackLogger()

        @rpc
        def method(self):
            return 1

    class Parent(object):
        name = "parent"

        stack_logger = StackLogger()
        child_service = ServiceRpc('child')

        @rpc
        def method(self):
            return self.child_service.method()

    class Grandparent(object):
        name = "grandparent"

        stack_logger = StackLogger()
        parent_service = ServiceRpc('parent')

        @rpc
        def method(self):
            return self.parent_service.method()

    runner = runner_factory()
    runner.add_service(Child)
    runner.add_service(Parent)
    runner.add_service(Grandparent)
    runner.start()

    container = get_container(runner, Grandparent)
    with entrypoint_hook(container, "method") as grandparent_method:
        assert grandparent_method() == 1

    # Check IDs were requested
    assert predictable_call_ids.call_count == 3

    # Check call ID stack persisted over RPC
    assert tracker.call_args_list == [
        call(['grandparent.method.0']),
        call(['grandparent.method.0', 'parent.method.1']),
        call(['grandparent.method.0', 'parent.method.1', 'child.method.2']),
    ]


def test_call_id_over_events(
    rabbit_config, predictable_call_ids, runner_factory, stack_logger, tracker
):
    StackLogger = stack_logger

    one_called = Mock()
    two_called = Mock()

    class EventListeningServiceOne(object):
        name = "listener_one"

        stack_logger = StackLogger()

        @event_handler('event_raiser', 'hello')
        def hello(self, name):
            one_called()

    class EventListeningServiceTwo(object):
        name = "listener_two"

        stack_logger = StackLogger()

        @event_handler('event_raiser', 'hello')
        def hello(self, name):
            two_called()

    class EventRaisingService(object):
        name = "event_raiser"
        dispatch = EventDispatcher()

        stack_logger = StackLogger()

        @rpc
        def say_hello(self):
            self.dispatch('hello', self.name)

    runner = runner_factory()
    runner.add_service(EventListeningServiceOne)
    runner.add_service(EventListeningServiceTwo)
    runner.add_service(EventRaisingService)
    runner.start()

    container = get_container(runner, EventRaisingService)
    listener1 = get_container(runner, EventListeningServiceOne)
    listener2 = get_container(runner, EventListeningServiceTwo)
    with entrypoint_hook(container, "say_hello") as say_hello:
        waiter1 = entrypoint_waiter(listener1, 'hello')
        waiter2 = entrypoint_waiter(listener2, 'hello')
        with waiter1, waiter2:
            say_hello()

    assert predictable_call_ids.call_count == 3

    # order of event handlers and dependencies is non-deterministic,
    # so there are four permutations of valid call stacks
    possible_call_lists = (
        [
            call(['event_raiser.say_hello.0']),
            call(['event_raiser.say_hello.0', 'listener_one.hello.1']),
            call(['event_raiser.say_hello.0', 'listener_two.hello.2']),
        ],
        [
            call(['event_raiser.say_hello.0']),
            call(['event_raiser.say_hello.0', 'listener_one.hello.2']),
            call(['event_raiser.say_hello.0', 'listener_two.hello.1']),
        ],
        [
            call(['event_raiser.say_hello.0']),
            call(['event_raiser.say_hello.0', 'listener_two.hello.1']),
            call(['event_raiser.say_hello.0', 'listener_one.hello.2']),
        ],
        [
            call(['event_raiser.say_hello.0']),
            call(['event_raiser.say_hello.0', 'listener_two.hello.2']),
            call(['event_raiser.say_hello.0', 'listener_one.hello.1']),
        ]
    )
    assert tracker.call_args_list in possible_call_lists


class TestImmediateParentCallId(object):

    def test_with_parent(self, mock_container):

        mock_container.service_name = "foo"

        service = Mock()
        entrypoint = DummyProvider("bar")
        context_data = {
            'call_id_stack': [
                'parent.method.1',
                'parent.method.2',
                'parent.method.3'
            ]
        }

        worker_ctx = WorkerContext(
            mock_container, service, entrypoint, data=context_data
        )

        assert worker_ctx.immediate_parent_call_id == "parent.method.3"

    def test_without_parent(self, mock_container):

        mock_container.service_name = "foo"

        service = Mock()
        entrypoint = DummyProvider("bar")
        context_data = {}

        worker_ctx = WorkerContext(
            mock_container, service, entrypoint, data=context_data
        )

        assert worker_ctx.immediate_parent_call_id is None


class TestOriginCallId(object):

    def test_with_origin(self, mock_container):

        mock_container.service_name = "foo"

        service = Mock()
        entrypoint = DummyProvider("bar")
        context_data = {
            'call_id_stack': [
                'parent.method.1',
                'parent.method.2',
                'parent.method.3'
            ]
        }

        worker_ctx = WorkerContext(
            mock_container, service, entrypoint, data=context_data
        )

        assert worker_ctx.origin_call_id == "parent.method.1"

    def test_without_origin(self, mock_container):

        mock_container.service_name = "foo"

        service = Mock()
        entrypoint = DummyProvider("bar")
        context_data = {}

        worker_ctx = WorkerContext(
            mock_container, service, entrypoint, data=context_data
        )

        assert worker_ctx.origin_call_id is None



import sys
from contextlib import contextmanager
from threading import Semaphore

import six
from mock import patch


class WaitResult(object):
    sentinel = object()

    res = sentinel
    exc_info = None

    class NotReady(Exception):
        pass

    @property
    def has_result(self):
        return (
            self.res is not self.sentinel or
            self.exc_info is not None
        )

    def send(self, res, exc_info):
        if not self.has_result:
            self.res = res
            self.exc_info = exc_info

    def get(self):
        if not self.has_result:
            raise WaitResult.NotReady()

        if self.exc_info is not None:
            six.reraise(*self.exc_info)
        return self.res


@contextmanager
def wait_for_call(obj, target, callback=None):

    sem = Semaphore(0)
    result = WaitResult()

    unpatched = getattr(obj, target)

    def maybe_release(args, kwargs, res, exc_info):
        should_release = True
        if callable(callback):
            should_release = callback(args, kwargs, res, exc_info)

        if should_release:
            result.send(res, exc_info)
            sem.release()

    def wraps(*args, **kwargs):
        res = None
        exc_info = None
        try:
            res = unpatched(*args, **kwargs)
        except Exception:
            exc_info = sys.exc_info()

        maybe_release(args, kwargs, res, exc_info)

        if exc_info is not None:
            six.reraise(*exc_info)
        return res

    with patch.object(obj, target, new=wraps):
        yield result
        sem.acquire()



import json
from nameko.web.handlers import HttpRequestHandler
from werkzeug.wrappers import Response
from nameko.exceptions import safe_for_serialization


class HttpError(Exception):
    error_code = 'BAD_REQUEST'
    status_code = 400


class InvalidArgumentsError(HttpError):
    error_code = 'INVALID_ARGUMENTS'


class HttpEntrypoint(HttpRequestHandler):
    def response_from_exception(self, exc):
        if isinstance(exc, HttpError):
            response = Response(
                json.dumps({
                    'error': exc.error_code,
                    'message': safe_for_serialization(exc),
                }),
                status=exc.status_code,
                mimetype='application/json'
            )
            return response
        return HttpRequestHandler.response_from_exception(self, exc)


http = HttpEntrypoint.decorator


class Service:
    name = "http_service"

    @http('GET', '/custom_exception')
    def custom_exception(self, request):
        raise InvalidArgumentsError("Argument `foo` is required.")



import pytest
from mock import Mock, call

from nameko import config
from nameko.events import BROADCAST, event_handler
from nameko.rpc import rpc
from nameko.runners import ServiceRunner, run_services
from nameko.standalone.events import event_dispatcher
from nameko.standalone.rpc import ServiceRpcClient
from nameko.testing.services import entrypoint_waiter
from nameko.testing.utils import assert_stops_raising, get_container


class TestService1(object):
    name = 'foobar_1'


class TestService2(object):
    name = 'foobar_2'


@pytest.fixture
def tracker():
    return Mock()


@pytest.fixture
def service_cls(tracker):

    class Service(object):
        name = "service"

        @rpc
        @event_handler(
            "srcservice", "testevent",
            handler_type=BROADCAST, reliable_delivery=False
        )
        def handle(self, msg):
            tracker(msg)

    return Service


@config.patch({'SERVICE_CONTAINER_CLS': 'fake_module.ServiceContainer'})
def test_runner_lifecycle(fake_module):
    events = set()

    class Container(object):
        def __init__(self, service_cls):
            self.service_name = service_cls.__name__
            self.service_cls = service_cls

        def start(self):
            events.add(('start', self.service_cls.name, self.service_cls))

        def stop(self):
            events.add(('stop', self.service_cls.name, self.service_cls))

        def kill(self):
            events.add(('kill', self.service_cls.name, self.service_cls))

        def wait(self):
            events.add(('wait', self.service_cls.name, self.service_cls))

    fake_module.ServiceContainer = Container

    runner = ServiceRunner()

    runner.add_service(TestService1)
    runner.add_service(TestService2)

    runner.start()

    assert events == {
        ('start', 'foobar_1', TestService1),
        ('start', 'foobar_2', TestService2),
    }

    events = set()
    runner.stop()
    assert events == {
        ('stop', 'foobar_1', TestService1),
        ('stop', 'foobar_2', TestService2),
    }

    events = set()
    runner.kill()
    assert events == {
        ('kill', 'foobar_1', TestService1),
        ('kill', 'foobar_2', TestService2),
    }

    events = set()
    runner.wait()
    assert events == {
        ('wait', 'foobar_1', TestService1),
        ('wait', 'foobar_2', TestService2),
    }


@config.patch({'SERVICE_CONTAINER_CLS': 'fake_module.ServiceContainer'})
def test_contextual_lifecycle(fake_module):
    events = set()

    class Container(object):
        def __init__(self, service_cls):
            self.service_name = service_cls.__name__
            self.service_cls = service_cls

        def start(self):
            events.add(('start', self.service_cls.name, self.service_cls))

        def stop(self):
            events.add(('stop', self.service_cls.name, self.service_cls))

        def kill(self, exc=None):
            events.add(('kill', self.service_cls.name, self.service_cls))

    fake_module.ServiceContainer = Container

    with run_services(TestService1, TestService2):
        # Ensure the services were started
        assert events == {
            ('start', 'foobar_1', TestService1),
            ('start', 'foobar_2', TestService2),
        }

    # ...and that they were stopped
    assert events == {
        ('start', 'foobar_1', TestService1),
        ('start', 'foobar_2', TestService2),
        ('stop', 'foobar_1', TestService1),
        ('stop', 'foobar_2', TestService2),
    }

    events = set()
    with run_services(TestService1, TestService2, kill_on_exit=True):
        # Ensure the services were started
        assert events == {
            ('start', 'foobar_1', TestService1),
            ('start', 'foobar_2', TestService2),
        }

    # ...and that they were killed
    assert events == {
        ('kill', 'foobar_1', TestService1),
        ('kill', 'foobar_2', TestService2),
        ('start', 'foobar_1', TestService1),
        ('start', 'foobar_2', TestService2),
    }


@config.patch({'SERVICE_CONTAINER_CLS': 'fake_module.ServiceContainer'})
def test_runner_waits_raises_error(fake_module):
    class Container(object):
        def __init__(self, service_cls):
            pass

        def start(self):
            pass

        def stop(self):
            pass

        def wait(self):
            raise Exception('error in container')

    fake_module.ServiceContainer = Container

    runner = ServiceRunner()
    runner.add_service(TestService1)
    runner.start()

    with pytest.raises(Exception) as exc_info:
        runner.wait()
    assert exc_info.value.args == ('error in container',)


@pytest.mark.usefixtures("rabbit_config")
def test_multiple_runners_coexist(
    runner_factory, get_vhost, rabbit_manager, service_cls, tracker
):

    runner1 = runner_factory(service_cls)
    runner1.start()

    runner2 = runner_factory(service_cls)
    runner2.start()

    vhost = get_vhost(config['AMQP_URI'])

    # verify there are two event queues with a single consumer each
    def check_consumers():
        evt_queues = [queue for queue in rabbit_manager.get_queues(vhost)
                      if queue['name'].startswith('evt-srcservice-testevent')]
        assert len(evt_queues) == 2
        for queue in evt_queues:
            assert queue['consumers'] == 1

    # rabbit's management API seems to lag
    assert_stops_raising(check_consumers)

    # test events (both services will receive if in "broadcast" mode)
    event_data = "event"
    dispatch = event_dispatcher()

    container1 = list(runner1.containers)[0]
    container2 = list(runner2.containers)[0]

    with entrypoint_waiter(container1, "handle"):
        with entrypoint_waiter(container2, "handle"):
            dispatch('srcservice', "testevent", event_data)
    assert tracker.call_args_list == [call(event_data), call(event_data)]

    # verify there are two consumers on the rpc queue
    rpc_queue = rabbit_manager.get_queue(vhost, 'rpc-service')
    assert rpc_queue['consumers'] == 2

    # test rpc (only one service will respond)
    arg = "arg"
    with ServiceRpcClient('service') as client:
        client.handle(arg)

    assert tracker.call_args_list == [
        call(event_data), call(event_data), call(arg)
    ]


@pytest.mark.usefixtures("rabbit_config")
def test_runner_with_duplicate_services(
    runner_factory, service_cls, tracker
):

    # host Service multiple times
    runner = runner_factory()
    runner.add_service(service_cls)
    runner.add_service(service_cls)  # no-op
    runner.start()

    # it should only be hosted once
    assert len(runner.containers) == 1
    container = list(runner.containers)[0]

    # test events (only one service is hosted)
    event_data = "event"
    dispatch = event_dispatcher()

    with entrypoint_waiter(container, "handle"):
        dispatch('srcservice', "testevent", event_data)
    assert tracker.call_args_list == [call(event_data)]

    # test rpc
    arg = "arg"
    with ServiceRpcClient("service") as client:
        client.handle(arg)

    assert tracker.call_args_list == [call(event_data), call(arg)]


@pytest.mark.usefixtures("rabbit_config")
def test_runner_catches_managed_thread_errors(runner_factory, service_cls):

    class Broken(Exception):
        pass

    def raises():
        raise Broken('error')

    runner = runner_factory(service_cls)

    container = get_container(runner, service_cls)
    container.spawn_managed_thread(raises)

    with pytest.raises(Broken):
        runner.wait()



import json

import six
from requests import ConnectionError, HTTPError, Session
from requests.auth import HTTPBasicAuth
from requests.utils import get_auth_from_url, urldefragauth
from six.moves.urllib.parse import quote  # pylint: disable=E0401


__all__ = ['Client', 'HTTPError']


def _quote(value):
    return quote(value, '')


class Client(object):
    """Pyrabbit replacement using requests instead of httplib2 """

    def __init__(self, uri):

        # move basic auth creds into headers to avoid
        # https://github.com/requests/requests/issues/4275
        username, password = get_auth_from_url(uri)
        uri = urldefragauth(uri)

        self._base_url = '{}/api'.format(uri)
        self._session = Session()
        self._session.auth = HTTPBasicAuth(username, password)
        self._session.headers['content-type'] = 'application/json'
        self._verify_api_connection()

    def _build_url(self, args):
        args = map(_quote, args)
        return '{}/{}'.format(
            self._base_url,
            '/'.join(args),
        )

    def _request(self, method, *args, **kwargs):
        url = self._build_url(args)
        json_data = kwargs.pop('json', None)
        if json_data is not None:
            kwargs['data'] = json.dumps(json_data)

        try:
            result = self._session.request(method, url, **kwargs)
        except ConnectionError as exc:
            six.raise_from(Exception(
                'Connection error for the RabbitMQ management HTTP'
                ' API at {}, is it enabled?'.format(url)
            ), exc)

        result.raise_for_status()
        if result.content:
            return result.json()

    def _get(self, *args, **kwargs):
        return self._request('GET', *args, **kwargs)

    def _put(self, *args, **kwargs):
        return self._request('PUT', *args, **kwargs)

    def _delete(self, *args, **kwargs):
        return self._request('DELETE', *args, **kwargs)

    def _post(self, *args, **kwargs):
        return self._request('POST', *args, **kwargs)

    def _verify_api_connection(self):
        self._get('overview')

    def get_connections(self):
        return self._get('connections')

    def get_exchanges(self, vhost):
        return self._get('exchanges', vhost)

    def get_all_vhosts(self):
        return self._get('vhosts')

    def create_vhost(self, vhost):
        return self._put('vhosts', vhost)

    def delete_vhost(self, vhost):
        return self._delete('vhosts', vhost)

    def set_vhost_permissions(self, vhost, username, configure, read, write):
        permissions = {
            'configure': configure,
            'read': read,
            'write': write,
        }
        return self._put(
            'permissions', vhost, username,
            json=permissions)

    def get_queue(self, vhost, name):
        return self._get('queues', vhost, name)

    def create_queue(self, vhost, name, **properties):
        return self._put('queues', vhost, name, json=properties)

    def get_queues(self, vhost):
        return self._get('queues', vhost)

    def get_queue_bindings(self, vhost, name):
        return self._get('queues', vhost, name, 'bindings')

    def create_queue_binding(self, vhost, exchange, queue, routing_key):
        body = {
            'routing_key': routing_key,
        }
        return self._post(
            'bindings', vhost, 'e', exchange, 'q', queue, json=body
        )

    def publish(self, vhost, name, routing_key, payload, properties=None):
        body = {
            'routing_key': routing_key,
            'payload': payload,
            'properties': properties or {},
            'payload_encoding': 'string',
        }
        return self._post('exchanges', vhost, name, 'publish', json=body)

    def get_messages(self, vhost, name, count=1, requeue=False):
        body = {
            'count': count,
            'encoding': 'auto',
            'requeue': requeue,
        }
        return self._post('queues', vhost, name, 'get', json=body)



from textwrap import dedent

import pytest
from mock import patch


@patch("nameko.cli.utils.config.os")
@pytest.mark.usefixtures("empty_config")
def test_main(mock_os, tmpdir, capsys, command):

    config_file = tmpdir.join("config.yaml")
    config_file.write(
        """
        FOO: ${FOO:foobar}
        BAR: ${BAR}
    """
    )

    mock_os.environ = {"BAR": "[1,2,3]"}

    command("nameko", "show-config", "--config", config_file.strpath)

    out, _ = capsys.readouterr()

    expected = dedent(
        """
        BAR:
        - 1
        - 2
        - 3
        FOO: foobar
    """
    ).strip()

    assert out.strip() == expected



# start a runner with a service that errors. does it hang? or stop?
# how do we get an individual servicecontainer to blow up?
import pytest
from mock import ANY, call, patch

from nameko.events import EventDispatcher
from nameko.exceptions import RemoteError
from nameko.rpc import RpcConsumer, ServiceRpc, rpc
from nameko.standalone.rpc import ServiceRpcClient
from nameko.testing.services import entrypoint_hook
from nameko.testing.utils import get_container, get_extension


class ExampleError(Exception):
    pass


class ExampleService(object):
    name = "exampleservice"

    dispatch = EventDispatcher()
    exampleservice_rpc = ServiceRpc('exampleservice')

    @rpc
    def task(self):
        return "task_result"

    @rpc
    def broken(self):
        raise ExampleError("broken")

    @rpc(expected_exceptions=ExampleError)
    def bad(self):
        raise ExampleError("bad")

    @rpc
    def proxy(self, method, *args):
        """ Makes RPC calls to ``method`` on itself, so we can test handling
        of errors in remote services.
        """
        getattr(self.exampleservice_rpc, method)(*args)


class SecondService(object):
    name = "secondservice"

    @rpc
    def task(self):
        return 'task_result'  # pragma: no cover


@pytest.mark.usefixtures("rabbit_config")
def test_error_in_worker(container_factory):

    container = container_factory(ExampleService)
    container.start()

    with patch('nameko.containers._log') as logger:
        with entrypoint_hook(container, "broken") as broken:
            with pytest.raises(ExampleError):
                broken()
    assert not container._died.ready()

    assert logger.exception.call_args == call(
        'error handling worker %s: %s', ANY, ANY)


@pytest.mark.usefixtures("rabbit_config")
def test_expected_error_in_worker(container_factory):

    container = container_factory(ExampleService)
    container.start()

    with patch('nameko.containers._log') as logger:
        with entrypoint_hook(container, "bad") as bad:
            with pytest.raises(ExampleError):
                bad()
    assert not container._died.ready()

    assert logger.warning.call_args == call(
        '(expected) error handling worker %s: %s', ANY, ANY, exc_info=True)


@pytest.mark.usefixtures("rabbit_config")
def test_error_in_remote_worker(container_factory):

    container = container_factory(ExampleService)
    container.start()

    with entrypoint_hook(container, "proxy") as proxy:
        with pytest.raises(RemoteError) as exc_info:
            proxy("broken")
    assert exc_info.value.exc_type == "ExampleError"
    assert not container._died.ready()


@pytest.mark.usefixtures("rabbit_config")
def test_handle_result_error(container_factory):

    container = container_factory(ExampleService)
    container.start()

    rpc_consumer = get_extension(container, RpcConsumer)
    with patch.object(
            rpc_consumer, 'handle_result', autospec=True
    ) as handle_result:
        err = "error in handle_result"
        handle_result.side_effect = Exception(err)

        # use a standalone rpc client to call exampleservice.task()
        with ServiceRpcClient("exampleservice") as client:
            # client.task() will hang forever because it generates an error
            client.task.call_async()

        with pytest.raises(Exception) as exc_info:
            container.wait()

        assert str(exc_info.value) == err


@pytest.mark.usefixtures("rabbit_config")
@pytest.mark.parametrize("method_name",
                         ["worker_setup", "worker_result", "worker_teardown"])
def test_dependency_call_lifecycle_errors(
        container_factory, rabbit_config, method_name):

    container = container_factory(ExampleService)
    container.start()

    dependency = get_extension(container, EventDispatcher)
    with patch.object(dependency, method_name, autospec=True) as method:
        err = "error in {}".format(method_name)
        method.side_effect = Exception(err)

        # use a standalone rpc client to call exampleservice.task()
        with ServiceRpcClient("exampleservice") as client:
            # client.task() will hang forever because it generates an error
            client.task.call_async()

        # verify that the error bubbles up to container.wait()
        with pytest.raises(Exception) as exc_info:
            container.wait()
        assert str(exc_info.value) == err


@pytest.mark.usefixtures("rabbit_config")
def test_runner_catches_container_errors(runner_factory):

    runner = runner_factory(ExampleService)
    runner.start()

    container = get_container(runner, ExampleService)

    rpc_consumer = get_extension(container, RpcConsumer)
    with patch.object(
            rpc_consumer, 'handle_result', autospec=True) as handle_result:
        exception = Exception("error")
        handle_result.side_effect = exception

        # use a standalone rpc client to call exampleservice.task()
        with ServiceRpcClient("exampleservice") as client:
            # client.task() will hang forever because it generates an error
            # in the remote container (so never receives a response).
            client.task.call_async()

        # verify that the error bubbles up to runner.wait()
        with pytest.raises(Exception) as exc_info:
            runner.wait()
        assert exc_info.value == exception


@pytest.mark.usefixtures("rabbit_config")
def test_graceful_stop_on_one_container_error(runner_factory):

    runner = runner_factory(ExampleService, SecondService)
    runner.start()

    container = get_container(runner, ExampleService)
    second_container = get_container(runner, SecondService)
    original_stop = second_container.stop
    with patch.object(second_container, 'stop', autospec=True,
                      wraps=original_stop) as stop:
        rpc_consumer = get_extension(container, RpcConsumer)
        with patch.object(
                rpc_consumer, 'handle_result', autospec=True) as handle_result:
            exception = Exception("error")
            handle_result.side_effect = exception

            # use a standalone rpc client to call exampleservice.task()
            with ServiceRpcClient("exampleservice") as client:
                # client.task() will hang forever because it generates an error
                # in the remote container (so never receives a response).
                client.task.call_async()

            # verify that the error bubbles up to runner.wait()
            with pytest.raises(Exception) as exc_info:
                runner.wait()
            assert exc_info.value == exception

            # Check that the second service was stopped due to the first
            # service being killed
            stop.assert_called_once_with()



import os
import sys

import pytest
import yaml
from mock import call, patch

from nameko.cli import cli
from nameko.cli.utils.config import ENV_VAR_MATCHER, setup_yaml_parser
from nameko.exceptions import CommandError, ConfigurationError


try:
    import regex
except ImportError:  # pragma: no cover
    has_regex_module = False
else:  # pragma: no cover
    del regex
    has_regex_module = True


@pytest.yield_fixture(autouse=True)
def fake_argv(empty_config):
    with patch.object(
        sys,
        "argv",
        [
            "nameko",
            "run",
            "--define",
            "AMQP_URI=pyamqp://someuser:*****@somehost/",
            "test.sample:Service",
        ],
    ):
        yield


def test_run():
    define = {"AMQP_URI": "pyamqp://someuser:*****@somehost/"}
    services = ("test.sample:Service",)
    config_file = None
    broker = None
    backdoor_port = None
    with patch("nameko.cli.run.main") as main_run:
        with patch("nameko.cli.setup_config") as setup_config:
            cli(standalone_mode=False)
            assert setup_config.call_args_list == [call(config_file, define, broker)]
            assert main_run.call_args_list == [call(services, backdoor_port)]


@pytest.mark.parametrize("exception", (CommandError, ConfigurationError))
def test_error(exception, capsys):
    with patch("nameko.cli.run.main") as main_run:
        with pytest.raises(SystemExit):
            main_run.side_effect = exception("boom")
            cli(standalone_mode=True)
    out, _ = capsys.readouterr()
    assert "Error: boom" in out.strip()


@pytest.mark.parametrize(
    ("text", "expected_key", "expected_value"),
    (
        # strings
        ("SPAM=ham", "SPAM", "ham"),
        ("  SPAM=ham  ", "SPAM", "ham"),
        ('SPAM="ham"', "SPAM", "ham"),
        ("SPAM='ham'", "SPAM", "ham"),
        ('SPAM="1"', "SPAM", "1"),
        # simple types
        ("SPAM=1", "SPAM", 1),
        ("SPAM=1.0", "SPAM", 1.0),
        ("SPAM=True", "SPAM", True),
        ("SPAM=False", "SPAM", False),
        ("SPAM", "SPAM", True),
        ("SPAM=", "SPAM", None),
        # structure
        ('SPAM=[{"egg": "ham"}]', "SPAM", [{"egg": "ham"}]),
        # deals with equal signs in value
        ("SPAM=EGG=HAM", "SPAM", "EGG=HAM"),
    ),
)
def test_parse_config_option(text, expected_key, expected_value):
    from nameko.cli.click_paramtypes import KeyValParamType

    assert KeyValParamType().convert(text, None, None) == (expected_key, expected_value)


class TestConfigEnvironmentParser(object):
    @pytest.mark.parametrize(
        ("value", "expected"),
        [
            # no var
            ("raw", []),
            # var without default
            ("${VAR_NAME}", [("VAR_NAME", "")]),
            # on var with default value
            ("${VAR_NAME:default}", [("VAR_NAME", "default")]),
            # multiple match with default
            (
                "${VAR_NAME1:default1}_${VAR_NAME2:default2}",
                [("VAR_NAME1", "default1"), ("VAR_NAME2", "default2")],
            ),
        ],
    )
    def test_maching_env_variable(self, value, expected):
        res = ENV_VAR_MATCHER.findall(value)
        assert res == expected

    @pytest.mark.parametrize(
        ("value", "expected"),
        [
            # on var, with {} in default (as str.format accept)
            (
                "${VAR_NAME:my-{pytemplate}-template}",
                [("VAR_NAME", "my-{pytemplate}-template")],
            ),
            # var with var in default
            (
                "${VAR_NAME:my-${OTHER_VAR}-template}",
                [("VAR_NAME", "my-${OTHER_VAR}-template")],
            ),
            # recursive interpretation of last var
            ("my-${OTHER_VAR}-template", [("OTHER_VAR", "")]),
            # var with var in default in default
            (
                "${VAR_NAME:my-${OTHER_VAR:1}-template}",
                [("VAR_NAME", "my-${OTHER_VAR:1}-template")],
            ),
            # recursive interpretation of last var
            ("my-${OTHER_VAR:1}-template", [("OTHER_VAR", "1")]),
            # multiple var in default
            (
                "${VAR_NAME1:default1}_${VAR_NAME2:default2}",
                [("VAR_NAME1", "default1"), ("VAR_NAME2", "default2")],
            ),
        ],
    )
    @pytest.mark.skipif(
        not has_regex_module, reason="0 support for nested env without regex module"
    )
    def test_maching_recursive_with_regex(self, value, expected):  # pragma: no cover
        res = ENV_VAR_MATCHER.findall(value)
        assert res == expected


class TestConfigEnvironmentVariables(object):
    @pytest.mark.parametrize(
        ("yaml_config", "env_vars", "expected_config"),
        [
            # no default value, no env value
            ("FOO: ${BAR}", {}, {"FOO": None}),
            # use default value if env value not provided
            ("FOO: ${BAR:foo}", {}, {"FOO": "foo"}),
            # use env value
            ("FOO: ${BAR}", {"BAR": "bar"}, {"FOO": "bar"}),
            # use env value instead of default
            ("FOO: ${BAR:foo}", {"BAR": "bar"}, {"FOO": "bar"}),
            # handles multi line
            (
                """
            FOO: ${BAR:foo}
            BAR: ${FOO:bar}
            """,
                {"BAR": "bar", "FOO": "foo"},
                {"FOO": "bar", "BAR": "foo"},
            ),
            # quoted values don't work without explicit resolver
            ('FOO: "${BAR:foo}"', {"BAR": "bar"}, {"FOO": "${BAR:foo}"}),
            # quoted values work only with explicit resolver
            ('FOO: !env_var "${BAR:foo}"', {"BAR": "bar"}, {"FOO": "bar"}),
            # quoted values with raw_env_var constructor to avoid yaml parsing
            ('FOO: !raw_env_var "${BAR}"', {"BAR": "123.0"}, {"FOO": "123.0"}),
            # $ sign can be used
            ("FOO: $bar", {}, {"FOO": "$bar"}),
            # multiple substitutions
            (
                "FOO: http://${BAR:foo}/${FOO:bar}",
                {"BAR": "bar", "FOO": "foo"},
                {"FOO": "http://bar/foo"},
            ),
            # can handle int, float and boolean
            (
                """
            FOO:
                - BAR: ${INT:1}
                - BAR: ${FLOAT:1.1}
                - BAR: ${BOOL:True}
            """,
                {},
                {"FOO": [{"BAR": 1}, {"BAR": 1.1}, {"BAR": True}]},
            ),
            (
                """
            FOO:
                - BAR: ${INT}
                - BAR: ${FLOAT}
                - BAR: ${BOOL}
            """,
                {"INT": "1", "FLOAT": "1.1", "BOOL": "True"},
                {"FOO": [{"BAR": 1}, {"BAR": 1.1}, {"BAR": True}]},
            ),
            # quoted default results in string
            (
                """
            FOO:
                - BAR: ${INT:"1"}
                - BAR: ${FLOAT:"1.1"}
                - BAR: ${BOOL:"True"}
            """,
                {},
                {"FOO": [{"BAR": "1"}, {"BAR": "1.1"}, {"BAR": "True"}]},
            ),
            # quoted value results in string
            # double-quote required
            (
                """
            FOO:
                - BAR: !env_var "'${INT}'"
                - BAR: !env_var "'${FLOAT}'"
                - BAR: !env_var "'${BOOL}'"
            """,
                {"INT": "1", "FLOAT": "1.1", "BOOL": "True"},
                {"FOO": [{"BAR": "1"}, {"BAR": "1.1"}, {"BAR": "True"}]},
            ),
            # quoted input results in string
            # double-quote required
            (
                """
            FOO:
                - BAR: ${INT}
                - BAR: ${FLOAT}
                - BAR: ${BOOL}
            """,
                {"INT": '"1"', "FLOAT": '"1.1"', "BOOL": '"True"'},
                {"FOO": [{"BAR": "1"}, {"BAR": "1.1"}, {"BAR": "True"}]},
            ),
            # list of scalar values
            (
                """
            FOO:
                - ${INT_1}
                - ${INT_2}
                - ${INT_3}
            BAR:
                - 1
                - 2
                - 3
            """,
                {"INT_1": "1", "INT_2": "2", "INT_3": "3"},
                {"FOO": [1, 2, 3], "BAR": [1, 2, 3]},
            ),
            # inline list
            (
                """
            FOO: ${LIST}
            BAR:
                - 1
                - 2
                - 3
            """,
                {"LIST": "[1,2,3]"},
                {"FOO": [1, 2, 3], "BAR": [1, 2, 3]},
            ),
            # inline list with block style
            (
                """
            FOO: ${LIST}
            BAR:
                - 1
                - 2
                - 3
            """,
                {"LIST": "- 1\n- 2\n- 3"},
                {"FOO": [1, 2, 3], "BAR": [1, 2, 3]},
            ),
            # inline list via explicit tag
            (
                """
            FOO: !env_var "${LIST}"
            BAR:
                - 1
                - 2
                - 3
            """,
                {"LIST": "[1,2,3]"},
                {"FOO": [1, 2, 3], "BAR": [1, 2, 3]},
            ),
            # inline list with default
            (
                """
            FOO: ${LIST:[1,2,3]}
            BAR:
                - 1
                - 2
                - 3
            """,
                {},
                {"FOO": [1, 2, 3], "BAR": [1, 2, 3]},
            ),
            # inline list containing list
            (
                """
            FOO: ${LIST}
            BAR:
                - 1
                - 2
                - 3
            """,
                {"LIST": "[1, 2, 3, ['a', 'b', 'c']]"},
                {"FOO": [1, 2, 3, ["a", "b", "c"]], "BAR": [1, 2, 3]},
            ),
            # inline dict
            (
                """
            FOO: ${DICT}
            BAR:
                - 1
                - 2
                - 3
            """,
                {"DICT": "{'one': 1, 'two': 2}"},
                {"FOO": {"one": 1, "two": 2}, "BAR": [1, 2, 3]},
            ),
            # inline dict with block style
            (
                """
            FOO: ${DICT}
            BAR:
                - 1
                - 2
                - 3
            """,
                {"DICT": "one: 1\ntwo: 2"},
                {"FOO": {"one": 1, "two": 2}, "BAR": [1, 2, 3]},
            ),
        ],
    )
    def test_environment_vars_in_config(self, yaml_config, env_vars, expected_config):
        setup_yaml_parser()

        with patch.dict("os.environ"):
            for key, val in env_vars.items():
                os.environ[key] = val

            results = yaml.unsafe_load(yaml_config)
            assert results == expected_config

    def test_cannot_recurse(self):

        setup_yaml_parser()

        yaml_config = """
            FOO: ${VAR1}
            BAR:
                - 1
                - 2
                - 3
        """

        with patch.dict("os.environ"):
            os.environ["VAR1"] = "${VAR1}"

            results = yaml.unsafe_load(yaml_config)
            assert results == {"FOO": "${VAR1}", "BAR": [1, 2, 3]}

    @pytest.mark.parametrize(
        ("yaml_config", "env_vars", "expected_config"),
        [
            # recursive env with root value
            (
                """
            FOO: ${FOO:val_${INDICE:1}}
            """,
                {"FOO": "val_a"},
                {"FOO": "val_a"},
            ),
            # recursive env with root default and sub value
            (
                """
            FOO: ${FOO:val_${INDICE:1}}
            """,
                {"INDICE": "b"},
                {"FOO": "val_b"},
            ),
            # recursive env requiring data
            (
                """
            FOO: ${FOO:val_${INDICE}}
            """,
                {"INDICE": "b"},
                {"FOO": "val_b"},
            ),
            # default data with bracket
            (
                """
            FOO: ${FOO:"{name} {age}"}
            """,
                {},
                {"FOO": "{name} {age}"},
            ),
            # default data with bracket
            (
                """
            FOO: ${FOO:"{name} {age}"}
            """,
                {"FOO": '"{surname}"'},
                {"FOO": "{surname}"},
            ),
        ],
    )
    @pytest.mark.skipif(
        not has_regex_module, reason="0 support for nested env without regex module"
    )
    def test_environment_vars_recursive_in_config(
        self, yaml_config, env_vars, expected_config
    ):  # pragma: no cover
        setup_yaml_parser()

        with patch.dict("os.environ"):
            for key, val in env_vars.items():
                os.environ[key] = val

            results = yaml.unsafe_load(yaml_config)
            assert results == expected_config

    @pytest.mark.parametrize(
        ("yaml_config", "env_vars", "expected_config"),
        [
            # recursive env with root value
            (
                """
            FOO: ${FOO:val_${INDICE:1}}
            """,
                {"FOO": "val_a"},
                {"FOO": "val_a}"},
            ),
            # recursive env with root default and sub value
            (
                """
            FOO: ${FOO:val_${INDICE:1}}
            """,
                {"INDICE": "b"},
                {"FOO": "val_${INDICE:1}"},
            ),
            # recursive env requiring data
            (
                """
            FOO: ${FOO:${INDICE}_val}
            """,
                {"INDICE": "b"},
                {"FOO": "${INDICE_val}"},
            ),
        ],
    )
    @pytest.mark.skipif(has_regex_module, reason="default behavior if no regex module")
    def test_unhandled_recursion(
        self, yaml_config, env_vars, expected_config
    ):  # pragma: no cover
        setup_yaml_parser()

        with patch.dict("os.environ"):
            for key, val in env_vars.items():
                os.environ[key] = val
            results = yaml.unsafe_load(yaml_config)
            assert results == expected_config



# websocket_rpc.py

from nameko.web.websocket import rpc, WebSocketHubProvider

class WebsocketRpc:
    name = "websocket_rpc_service"

    websocket_hub = WebSocketHubProvider()

    @rpc
    def echo(self, socket_id, value):
        return value



from nameko.extensions import DependencyProvider

import boto3

class SqsSend(DependencyProvider):

    def __init__(self, url, region="eu-west-1", **kwargs):
        self.url = url
        self.region = region
        super(SqsSend, self).__init__(**kwargs)

    def setup(self):
        self.client = boto3.client('sqs', region_name=self.region)

    def get_dependency(self, worker_ctx):

        def send_message(payload):
            # assumes boto client is thread-safe for this action, because it
            # happens inside the worker threads
            self.client.send_message(
                QueueUrl=self.url,
                MessageBody=payload
            )
        return send_message



import eventlet
import pytest
from kombu.messaging import Queue
from mock import Mock, patch

from nameko.amqp.consume import Consumer
from nameko.amqp.publish import Publisher


class TestConsumer(object):

    @pytest.fixture
    def queue(self):
        return Queue(name="queue")

    @pytest.fixture
    def publisher(self, amqp_uri):
        return Publisher(amqp_uri)

    def test_consume_from_queue(self, queue, publisher, amqp_uri):

        tracker = Mock()

        def callback(body, message):
            tracker(body)
            message.ack()

        publisher.publish("payload", routing_key=queue.name, declare=[queue])

        consumer = Consumer(
            amqp_uri, queues=[queue], callbacks=[callback]
        )
        next(consumer.consume())

        assert tracker.call_count == 1

    def test_wait_until_consumer_ready(self, queue, publisher, amqp_uri):

        consumer = Consumer(amqp_uri, queues=[queue])

        def waiter():
            consumer.wait_until_consumer_ready()

        gt = eventlet.spawn(waiter)

        assert not consumer.ready.is_set()
        assert not gt.dead

        publisher.publish("payload", routing_key=queue.name, declare=[queue])
        next(consumer.consume())

        assert consumer.ready.is_set()
        gt.wait()  # make sure gt gets scheduled and has chance to exit
        assert gt.dead

    def test_ack_message_ignores_connection_errors(
        self, queue, publisher, amqp_uri
    ):
        publisher.publish("payload", routing_key=queue.name, declare=[queue])

        messages = []

        consumer = Consumer(
            amqp_uri,
            queues=[queue],
            callbacks=[lambda body, message: messages.append(message)]
        )
        next(consumer.consume())

        consumer.connection.close()

        for message in messages:
            with patch.object(message, 'ack') as ack:
                consumer.ack_message(message)
            assert not ack.called

    def test_requeue_message_ignores_connection_errors(
        self, queue, publisher, amqp_uri
    ):
        publisher.publish("payload", routing_key=queue.name, declare=[queue])

        messages = []

        consumer = Consumer(
            amqp_uri,
            queues=[queue],
            callbacks=[lambda body, message: messages.append(message)]
        )
        next(consumer.consume())

        consumer.connection.close()

        for message in messages:
            with patch.object(message, 'requeue') as requeue:
                consumer.requeue_message(message)
            assert not requeue.called

    def test_messages_requeued_after_stop(self, amqp_uri):

        messages = []

        consumer = Consumer(
            amqp_uri,
            callbacks=[lambda body, message: messages.append(message)]
        )

        message = Mock()

        consumer.on_message("body", message)

        assert len(messages) == 1
        assert not message.requeue.called

        consumer.stop()

        consumer.on_message("body", message)

        assert len(messages) == 1  # still 1, callback not called
        assert message.requeue.called



from nameko.standalone.rpc import ClusterRpcClient


with ClusterRpcClient(
    uri=AMQP_URI  # e.g. "pyamqp://guest:guest@localhost"
) as cluster_rpc:
    cluster_rpc.service_x.remote_method("hellø")  # "hellø-x-y"



""" DEPRECATED Nameko built-in dependencies.
"""
import warnings

from nameko import config
from nameko.extensions import DependencyProvider


class Config(DependencyProvider):
    """ DEPRECATED Dependency provider for accessing configuration values.
    """

    def get_dependency(self, worker_ctx):
        warnings.warn("Use ``nameko.config`` instead.", DeprecationWarning)
        return config.copy()



"""Utility function `import_services` to import nameko services
from strings pointing either to modules (then provide all nameko services
present there) or to explicit nameko service class.

Note: If you want to load locally present python files as modules,
modify sys.path::

    import sys
    if not "." in sys.path:
        sys.path.insert(0, ".")
    services = import_services("my_service")  # load my_service.py

See also NamekoModuleServicesParamType if you want to use it with click.
"""
import inspect
import os
import re
import sys

import six

from nameko.extensions import ENTRYPOINT_EXTENSIONS_ATTR


MISSING_MODULE_TEMPLATE = "^No module named '?{}'?$"


def is_class(obj):
    return isinstance(obj, six.class_types)


def is_entrypoint(method):
    return hasattr(method, ENTRYPOINT_EXTENSIONS_ATTR)


def import_services(module_name):
    """import nameko services from specified module (with optional service name)

    `module_name` value can be in style:
    - 'amodule': all nameko services in module amodule
    - 'pck.amodule': all nameko services in module 'pck.amodule'
    - 'amodule:MySvc': 'MySvc' nameko class from amodule
    - 'pck.amodule:MySvc': 'MySvc' nameko service from module 'pck.amodule'

    Raises:

    - ValueError with explanatory message and possible hint. Covers scenarios
      such as: missing module, no nameko class found in the module, explicit
      nameko service class is missing, is not a class or is not nameko service.
    - In case of complex failure, the ValueError string will include stacktrace.

    Returns: list of imported services
    """
    parts = module_name.split(":", 1)
    if len(parts) == 1:
        module_name, service_name = module_name, None
    else:
        module_name, service_name = parts[0], parts[1]

    try:
        __import__(module_name)
    except ImportError as exc:
        if module_name.endswith(".py") and os.path.exists(module_name):
            raise ValueError(
                "Failed to find service, did you mean '{}'?".format(
                    module_name[:-3].replace("/", ".")
                )
            )

        missing_module_re = MISSING_MODULE_TEMPLATE.format(module_name)
        # is there a better way to do this?

        if re.match(missing_module_re, str(exc)):
            raise ValueError(exc)

        # found module, but importing it raised an import error elsewhere
        # let this bubble (resulting in a full stacktrace being printed)
        raise ValueError(str(exc))

    module = sys.modules[module_name]

    if service_name is None:
        found_services = []
        # find top-level objects with entrypoints
        for _, potential_service in inspect.getmembers(module, is_class):
            if inspect.getmembers(potential_service, is_entrypoint):
                found_services.append(potential_service)

        if not found_services:
            raise ValueError(
                "Failed to find anything that looks like a service in module "
                "{!r}".format(module_name)
            )

    else:
        try:
            service_cls = getattr(module, service_name)
        except AttributeError:
            raise ValueError(
                "Failed to find service class {!r} in module {!r}".format(
                    service_name, module_name
                )
            )

        if not is_class(service_cls):
            raise ValueError("Service must be a class.")

        found_services = [service_cls]

    return found_services



""" Service unit testing best practice, with an alternative dependency.
"""

import pytest
from sqlalchemy import Column, Integer, String, create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from nameko.rpc import rpc
from nameko.testing.services import worker_factory

# using community extension from http://pypi.python.org/pypi/nameko-sqlalchemy
from nameko_sqlalchemy import Session


Base = declarative_base()


class Result(Base):
    __tablename__ = 'model'
    id = Column(Integer, primary_key=True)
    value = Column(String(64))


class Service:
    """ Service under test
    """
    name = "service"

    db = Session(Base)

    @rpc
    def save(self, value):
        result = Result(value=value)
        self.db.add(result)
        self.db.commit()


@pytest.fixture
def session():
    """ Create a test database and session
    """
    engine = create_engine('sqlite:///:memory:')
    Base.metadata.create_all(engine)
    session_cls = sessionmaker(bind=engine)
    return session_cls()


def test_service(session):

    # create instance, providing the test database session
    service = worker_factory(Service, db=session)

    # verify ``save`` logic by querying the test database
    service.save("helloworld")
    assert session.query(Result.value).all() == [("helloworld",)]



import pytest
from mock import Mock, patch

from nameko.utils.retry import retry


@pytest.fixture
def tracker():
    return Mock()


@pytest.yield_fixture(autouse=True)
def mock_sleep():
    with patch('nameko.utils.retry.sleep') as patched:

        def total():
            return sum(delay for (delay,), _ in patched.call_args_list)

        patched.total = total
        yield patched


def test_without_arguments(tracker, mock_sleep):

    @retry
    def fn():
        tracker()
        raise ValueError()

    with pytest.raises(ValueError):
        fn()

    assert tracker.call_count > 1


@pytest.mark.parametrize("exceptions", [ValueError, (ValueError,)])
def test_retry_for_exceptions(exceptions, tracker):
    threshold = 10

    @retry(for_exceptions=exceptions, max_attempts=float('inf'))
    def fn():
        tracker()
        if tracker.call_count < threshold:
            raise ValueError()
        else:
            raise KeyError()

    with pytest.raises(KeyError):
        fn()

    assert tracker.call_count == threshold


class TestRetries(object):

    def test_retry_limit(self, tracker):
        max_attempts = 5

        @retry(max_attempts=max_attempts)
        def fn():
            tracker()
            raise ValueError()

        with pytest.raises(ValueError):
            fn()

        assert tracker.call_count == 1 + max_attempts

    def test_retry_forever(self, tracker):
        threshold = 10

        @retry(max_attempts=None)
        def fn():
            tracker()
            if tracker.call_count == threshold:
                return threshold
            else:
                raise ValueError()

        assert fn() == threshold
        assert tracker.call_count == threshold


class TestDelay(object):

    def test_fixed_delay(self, tracker, mock_sleep):

        max_attempts = 5
        delay = 1

        @retry(max_attempts=max_attempts, delay=delay)
        def fn():
            tracker()
            raise ValueError()

        with pytest.raises(ValueError):
            fn()

        total_delay = mock_sleep.total()
        assert total_delay == delay * max_attempts

    def test_backoff(self, tracker, mock_sleep):

        max_attempts = 5
        delay = 2
        backoff = 3

        @retry(max_attempts=max_attempts, delay=delay, backoff=backoff)
        def fn():
            tracker()
            raise ValueError()

        with pytest.raises(ValueError):
            fn()

        total_delay = mock_sleep.total()
        assert total_delay == sum(
            delay * (backoff ** attempt)
            for attempt in range(1, max_attempts + 1)
        )

    def test_max_delay(self, tracker, mock_sleep):

        max_attempts = 5
        delay = 1
        backoff = 2
        max_delay = delay  # never increase delay, regardless of backoff

        @retry(
            max_attempts=max_attempts, delay=delay,
            backoff=backoff, max_delay=max_delay
        )
        def fn():
            tracker()
            raise ValueError()

        with pytest.raises(ValueError):
            fn()

        total_delay = mock_sleep.total()
        assert total_delay == delay * max_attempts



"""Custom click argument  decorators

Allow reuse of arguments across subcommands.

Reuse in other click based commands is also possible.

See nameko/cli/__init__.py for real use.
"""
from functools import partial

import click


argument_services = partial(
    click.argument,
    "services",
    nargs=-1,
    required=True,
    metavar="module[:service class]",
    type=str,
)



import pytest
from mock import ANY, Mock

from nameko.rpc import Responder


# python version compat
EXCEPTION_MODULE = Exception.__module__


@pytest.yield_fixture
def unserializable():
    def unserializable_inner():
        pass  # pragma: no cover
    yield unserializable_inner


@pytest.fixture
def message():
    message = Mock()
    message.properties = {'reply_to': '', 'content_type': 'application/json'}
    return message


def test_responder(message, mock_producer):

    exchange = Mock()

    responder = Responder('amqp://localhost', exchange, message)

    # serialisable result
    result, exc_info = responder.send_response(True, None)
    assert result is True
    assert exc_info is None

    expected_msg = {
        'result': True,
        'error': None
    }
    (msg,), _ = mock_producer.publish.call_args
    assert msg == expected_msg


def test_responder_worker_exc(message, mock_producer):

    exchange = Mock()

    responder = Responder('amqp://localhost', exchange, message)

    # serialisable exception
    worker_exc = Exception('error')
    result, exc_info = responder.send_response(
        None, (Exception, worker_exc, "tb"))
    assert result is None
    assert exc_info == (Exception, worker_exc, "tb")

    expected_msg = {
        'result': None,
        'error': {
            'exc_path': '{}.Exception'.format(EXCEPTION_MODULE),
            'value': 'error',
            'exc_type': 'Exception',
            'exc_args': ['error']
        }
    }
    (msg,), _ = mock_producer.publish.call_args
    assert msg == expected_msg


@pytest.mark.parametrize("content_type,exception_info_string", [
    ('application/json', "is not JSON serializable"),
    ('application/x-python-serialize', "Can't pickle")])
def test_responder_unserializable_result(
        message, mock_producer, unserializable,
        content_type, exception_info_string):

    message.properties['content_type'] = content_type

    exchange = Mock()

    responder = Responder('amqp://localhost', exchange, message)

    # unserialisable result
    worker_result = unserializable
    result, exc_info = responder.send_response(worker_result, None)

    # responder will return the error from the serializer
    assert result is None
    # Different kombu versions return different exceptions, so
    # testing for the concrete exception is not feasible
    assert exc_info == (ANY, ANY, ANY)
    assert exception_info_string in str(exc_info[1])

    # and publish a dictionary-serialized UnserializableValueError
    # on worker_result
    expected_msg = {
        'result': None,
        'error': {
            'exc_path': 'nameko.exceptions.UnserializableValueError',
            'value': 'Unserializable value: `{}`'.format(worker_result),
            'exc_type': 'UnserializableValueError',
            'exc_args': [],
        }
    }
    (msg,), _ = mock_producer.publish.call_args
    assert msg == expected_msg


def test_responder_cannot_unicode_exc(message, mock_producer):

    exchange = Mock()

    responder = Responder('amqp://localhost', exchange, message)

    class CannotUnicode(object):
        def __str__(self):
            raise Exception('error')

    # un-unicode-able exception
    worker_exc = Exception(CannotUnicode())

    # send_response should not throw
    responder.send_response(True, (Exception, worker_exc, "tb"))


def test_responder_cannot_repr_exc(message, mock_producer):

    exchange = Mock()

    responder = Responder('amqp://localhost', exchange, message)

    class CannotRepr(object):
        def __repr__(self):
            raise Exception('error')

    # un-repr-able exception
    worker_exc = Exception(CannotRepr())

    # send_response should not throw
    responder.send_response(True, (Exception, worker_exc, "tb"))



import os
import subprocess
import sys

import pytest
from mock import Mock, patch

from nameko.cli.shell import make_nameko_helper
from nameko.constants import (
    AMQP_URI_CONFIG_KEY, SERIALIZER_CONFIG_KEY, WEB_SERVER_CONFIG_KEY
)
from nameko.rpc import Client


def test_helper_module(rabbit_config):
    helper = make_nameko_helper()
    assert isinstance(helper.rpc, Client)
    helper.disconnect()


@pytest.yield_fixture
def pystartup(tmpdir):
    startup = tmpdir.join("startup.py")
    startup.write("foo = 42")

    with patch.dict(os.environ, {"PYTHONSTARTUP": str(startup)}):
        yield


@pytest.yield_fixture(autouse=True)
def fake_alternative_interpreters():
    # Make sure these appear unavailable even if installed. We cheat slightly,
    # and have the call to `embed` raise the ImportError, rather than the
    # actual module import (this is easier to do and has the same effect in our
    # case).
    fake_module = Mock()
    fake_module.embed.side_effect = ImportError
    with patch.dict(sys.modules, {"IPython": fake_module, "bpython": fake_module}):
        yield


@pytest.fixture
def isatty():
    with patch("nameko.cli.shell.sys.stdin.isatty") as isatty:
        yield isatty


@pytest.mark.usefixtures("rabbit_config")
def test_basic(command, pystartup):
    with patch("nameko.cli.shell.interact") as interact:
        command('nameko', 'shell')

    _, kwargs = interact.call_args
    local = kwargs["local"]
    assert "n" in local.keys()
    assert local["foo"] == 42
    local["n"].disconnect()


@pytest.mark.usefixtures("rabbit_config")
def test_plain(command, pystartup):
    with patch("nameko.cli.shell.interact") as interact:
        command("nameko", "shell", "--interface", "plain")

    _, kwargs = interact.call_args
    local = kwargs["local"]
    assert "n" in local.keys()
    assert local["foo"] == 42
    local["n"].disconnect()


@pytest.mark.usefixtures("rabbit_config")
def test_plain_fallback(command, pystartup):
    with patch("nameko.cli.shell.interact") as interact:
        command("nameko", "shell", "--interface", "bpython")

    _, kwargs = interact.call_args
    local = kwargs["local"]
    assert "n" in local.keys()
    assert local["foo"] == 42
    local["n"].disconnect()


@pytest.mark.usefixtures("rabbit_config")
def test_bpython(command, pystartup, isatty):

    isatty.return_value = True

    with patch("bpython.embed") as embed:
        command("nameko", "shell", "--interface", "bpython")

    _, kwargs = embed.call_args
    local = kwargs["locals_"]
    assert "n" in local.keys()
    assert local["foo"] == 42
    local["n"].disconnect()


@pytest.mark.usefixtures("rabbit_config")
def test_ipython(command, pystartup, isatty):

    isatty.return_value = True

    with patch("IPython.embed") as embed:
        command("nameko", "shell", "--interface", "ipython")

    _, kwargs = embed.call_args
    local = kwargs["user_ns"]
    assert "n" in local.keys()
    assert local["foo"] == 42
    local["n"].disconnect()


@pytest.mark.usefixtures("rabbit_config")
def test_uses_plain_when_not_tty(command, pystartup, isatty):

    isatty.return_value = False

    with patch("nameko.cli.shell.interact") as interact:
        command("nameko", "shell", "--interface", "ipython")

    assert interact.called


@pytest.mark.usefixtures("rabbit_config")
def test_config(command, pystartup, tmpdir):

    config_file = tmpdir.join("config.yaml")
    config_file.write(
        """
        WEB_SERVER_ADDRESS: '0.0.0.0:8001'
        serializer: 'json'
    """
    )

    with patch("nameko.cli.shell.interact") as interact:
        command("nameko", "shell", "--config", config_file.strpath)

    _, kwargs = interact.call_args
    local = kwargs["local"]
    assert "n" in local.keys()
    assert local["n"].config[WEB_SERVER_CONFIG_KEY] == "0.0.0.0:8001"
    assert local["n"].config[SERIALIZER_CONFIG_KEY] == "json"
    local["n"].disconnect()


@pytest.mark.usefixtures("rabbit_config")
def test_config_options(command, pystartup, tmpdir):

    config_file = tmpdir.join("config.yaml")
    config_file.write(
        """
        WEB_SERVER_ADDRESS: '0.0.0.0:8001'
        serializer: 'json'
    """
    )

    with patch("nameko.cli.shell.interact") as interact:
        command(
            "nameko",
            "shell",
            "--config",
            config_file.strpath,
            "--define",
            "serializer=pickle",
            "--define",
            'EGG=[{"spam": True}]',
        )

    _, kwargs = interact.call_args
    local = kwargs["local"]
    assert "n" in local.keys()
    assert local["n"].config[WEB_SERVER_CONFIG_KEY] == "0.0.0.0:8001"
    assert local["n"].config[SERIALIZER_CONFIG_KEY] == "pickle"
    assert local["n"].config["EGG"] == [{"spam": True}]
    local["n"].disconnect()


@pytest.mark.usefixtures("empty_config")
class TestBanner(object):
    @pytest.yield_fixture(autouse=True)
    def patch_nameko_helper(self):
        with patch("nameko.cli.shell.make_nameko_helper"):
            yield

    def test_banner(self, command, tmpdir):

        amqp_uri = "amqp://broker/config"

        config_file = tmpdir.join("config.yaml")
        config_file.write(
            """
            WEB_SERVER_ADDRESS: '0.0.0.0:8001'
            AMQP_URI: '{}'
            serializer: 'json'
        """.format(
                amqp_uri
            )
        )

        with patch("nameko.cli.shell.ShellRunner") as shell_runner:
            command("nameko", "shell", "--config", config_file.strpath)

        expected_message = "Broker: {}".format(amqp_uri)
        (banner, _), _ = shell_runner.call_args
        assert expected_message in banner


@pytest.mark.usefixtures("empty_config")
def test_broker_option_deprecated(command, rabbit_uri):

    with patch("nameko.cli.shell.interact") as interact:
        with patch("nameko.cli.utils.config.warnings") as warnings:
            command("nameko", "shell", "--broker", rabbit_uri)

    _, kwargs = interact.call_args
    local = kwargs["local"]
    assert "n" in local.keys()
    assert local["n"].config[AMQP_URI_CONFIG_KEY] == rabbit_uri
    local["n"].disconnect()

    assert warnings.warn.call_count


class TestExitCode:
    def test_exit_code_0(self):
        """ Test ``echo print(1) | nameko shell``

        This command should return an exit code of 0.
        """

        echo_process = subprocess.Popen(("echo", "print(1)"), stdout=subprocess.PIPE)
        nameko_process = subprocess.Popen(
            ("nameko", "shell", "--define=AMQP_URI=memory://localhost"),
            stdin=echo_process.stdout,
        )
        nameko_process.wait()
        assert nameko_process.returncode == 0

    def test_exit_code_1(self):
        """ Test ``echo raise Exception() | nameko shell``

        This command should return an exit code of 1.
        """

        echo_process = subprocess.Popen(
            ("echo", "raise Exception()"), stdout=subprocess.PIPE
        )
        nameko_process = subprocess.Popen(
            ("nameko", "shell", "--define=AMQP_URI=memory://localhost"),
            stdin=echo_process.stdout,
        )
        nameko_process.wait()
        assert nameko_process.returncode == 1



""" Service integration testing best practice.
"""

import pytest

from nameko.rpc import ServiceRpc, rpc
from nameko.testing.services import entrypoint_hook
from nameko.testing.utils import get_container


class ServiceX:
    """ Service under test
    """
    name = "service_x"

    y = ServiceRpc("service_y")

    @rpc
    def remote_method(self, value):
        res = "{}-x".format(value)
        return self.y.append_identifier(res)


class ServiceY:
    """ Service under test
    """
    name = "service_y"

    @rpc
    def append_identifier(self, value):
        return "{}-y".format(value)


@pytest.mark.usefixtures("rabbit_config")
def test_service_x_y_integration(runner_factory):

    # run services in the normal manner
    runner = runner_factory(ServiceX, ServiceY)
    runner.start()

    # artificially fire the "remote_method" entrypoint on ServiceX
    # and verify response
    container = get_container(runner, ServiceX)
    with entrypoint_hook(container, "remote_method") as entrypoint:
        assert entrypoint("value") == "value-x-y"



from nameko.rpc import ServiceRpc, rpc


class ServiceY:
    name = "service_y"

    @rpc
    def append_identifier(self, value):
        return u"{}-y".format(value)


class ServiceX:
    name = "service_x"

    y = ServiceRpc("service_y")

    @rpc
    def remote_method(self, value):
        res = u"{}-x".format(value)
        return self.y.append_identifier(res)



import json
import subprocess
import sys
import uuid
from contextlib import contextmanager
from types import ModuleType

import eventlet
import pytest
import requests
from kombu.messaging import Queue
from mock import ANY, patch
from six.moves import queue
from six.moves.urllib.parse import urlparse

from nameko import config
from nameko.amqp.publish import get_connection
from nameko.testing.utils import find_free_port
from nameko.utils.retry import retry


TOXIPROXY_HOST = "127.0.0.1"
TOXIPROXY_PORT = 8474


def pytest_configure(config):
    config.addinivalue_line(
        "markers",
        "publish_retry: distinguish tests that should use retry in publishers"
    )
    config.addinivalue_line(
        "markers",
        "behavioural: distinguish behavioural tests"
    )


@pytest.yield_fixture
def memory_rabbit_config(rabbit_config):
    with config.patch({'AMQP_URI': 'memory://localhost'}):
        yield


@pytest.yield_fixture
def mock_producer():
    with patch('nameko.amqp.publish.producers') as patched:
        with patched[ANY].acquire() as producer:
            # normal behaviour is for no messages to be returned
            producer.channel.returned_messages.get_nowait.side_effect = (
                queue.Empty
            )
            yield producer


@pytest.yield_fixture
def mock_channel():
    with patch('nameko.amqp.publish.connections') as patched:
        with patched[ANY].acquire() as connection:
            yield connection.channel()


@pytest.yield_fixture(scope='session')
def toxiproxy_server():
    # start a toxiproxy server
    host = TOXIPROXY_HOST
    port = TOXIPROXY_PORT
    server = subprocess.Popen(
        ['toxiproxy-server', '-port', str(port), '-host', host]
    )

    class NotReady(Exception):
        pass

    @retry(delay=0.1, max_attempts=10)
    def wait_until_server_ready():
        url = "http://{}:{}/proxies".format(TOXIPROXY_HOST, TOXIPROXY_PORT)
        res = requests.get(url)
        if not res.status_code == 200:  # pragma: no cover
            raise NotReady("toxiproxy-server failed to start")

    wait_until_server_ready()
    yield "{}:{}".format(host, port)
    server.terminate()


@pytest.yield_fixture
def toxiproxy(toxiproxy_server, rabbit_config):
    """ Insert a toxiproxy in front of RabbitMQ

    https://github.com/douglas/toxiproxy-python is not released yet, so
    we use requests to control the server.
    """

    # extract rabbit connection details
    amqp_uri = config['AMQP_URI']
    uri = urlparse(amqp_uri)
    rabbit_port = uri.port

    proxy_port = find_free_port()

    # create proxy
    proxy_name = "nameko_test_rabbitmq_{}".format(uuid.uuid4().hex)

    listen = "{}:{}".format(uri.hostname, proxy_port)
    upstream = "{}:{}".format(uri.hostname, rabbit_port)
    requests.post(
        'http://{}/proxies'.format(toxiproxy_server),
        data=json.dumps({
            'name': proxy_name,
            'listen': listen,
            'upstream': upstream
        })
    )

    # create proxied uri for publisher
    proxy_uri = "{}://{}:{}@{}{}".format(
        uri.scheme, uri.username, uri.password, listen, uri.path
    )

    toxic_name = '{}_timeout'.format(proxy_name)

    class Controller(object):

        def __init__(self, proxy_uri):
            self.uri = proxy_uri

        def enable(self):
            resource = 'http://{}/proxies/{}'.format(
                toxiproxy_server, proxy_name
            )
            data = {
                'enabled': True
            }
            requests.post(resource, json.dumps(data))

        def disable(self):
            resource = 'http://{}/proxies/{}'.format(
                toxiproxy_server, proxy_name
            )
            data = {
                'enabled': False
            }
            requests.post(resource, json.dumps(data))

        def set_timeout(self, timeout=500, stream="upstream"):
            resource = 'http://{}/proxies/{}/toxics'.format(
                toxiproxy_server, proxy_name
            )
            data = {
                'name': toxic_name,
                'type': 'timeout',
                'stream': stream,
                'attributes': {
                    'timeout': timeout
                }
            }
            requests.post(resource, json.dumps(data))

        def reset_timeout(self):
            resource = 'http://{}/proxies/{}/toxics/{}'.format(
                toxiproxy_server, proxy_name, toxic_name
            )
            requests.delete(resource)

        @contextmanager
        def disabled(self):
            self.disable()
            yield
            self.enable()

        @contextmanager
        def timeout(self, timeout=500, stream="upstream"):
            self.set_timeout(timeout=timeout, stream=stream)
            yield
            self.reset_timeout()

    controller = Controller(proxy_uri)
    yield controller

    # delete proxy
    # allow some grace period to ensure we don't remove the proxy before
    # other fixtures have torn down
    resource = 'http://{}/proxies/{}'.format(toxiproxy_server, proxy_name)
    eventlet.spawn_after(10, requests.delete, resource)


@pytest.yield_fixture
def fake_module():
    module = ModuleType("fake_module")
    sys.modules[module.__name__] = module
    yield module
    del sys.modules[module.__name__]


@pytest.fixture
def add_to_sys_path():

    @contextmanager
    def modify_sys_path(path):
        sys.path.insert(0, path)
        yield
        sys.path.remove(path)

    return modify_sys_path


@pytest.fixture
def queue_info(amqp_uri):
    # TODO once https://github.com/nameko/nameko/pull/484 lands
    # we can use the utility in nameko.amqo.utils instead of this
    def get_queue_info(queue_name):
        with get_connection(amqp_uri) as conn:
            queue = Queue(name=queue_name)
            queue = queue.bind(conn)
            return queue.queue_declare(passive=True)
    return get_queue_info


@pytest.fixture
def get_vhost():
    def parse(uri):
        from six.moves.urllib.parse import urlparse  # pylint: disable=E0401
        uri_parts = urlparse(uri)
        return uri_parts.path[1:]
    return parse



#!/usr/bin/env python
import os
from codecs import open

from setuptools import find_packages, setup


here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.rst'), 'r', 'utf-8') as handle:
    readme = handle.read()


setup(
    name='nameko',
    version='v3.0.0-rc9',
    description='A microservices framework for Python that lets service '
                'developers concentrate on application logic and encourages '
                'testability.',
    long_description=readme,
    author='onefinestay',
    url='http://github.com/nameko/nameko',
    packages=find_packages(exclude=['test', 'test.*']),
    install_requires=[
        "click>=7.0",
        "dnspython<2",
        "eventlet>=0.20.1;python_version<'3.6'",
        "eventlet>=0.21.0;python_version>='3.6' and python_version<'3.7'",
        "eventlet>=0.26.0;python_version>='3.7'",
        "kombu>=4.2.0,<5",
        "mock>=1.2",
        "path.py>=6.2",
        "pyyaml>=5.1",
        "requests>=1.2.0",
        "six>=1.9.0",
        "werkzeug>=0.9",
        "wrapt>=1.0.0",
    ],
    extras_require={
        'dev': [
            "astroid==2.2.5;python_version>='3'",
            "astroid==1.6.5;python_version<'3'",
            "coverage==4.5.1",
            "flake8==3.3.0",
            "isort==4.2.15",
            "mccabe==0.6.1",
            "pycodestyle==2.3.1",
            "pyflakes==1.5.0",
            "pylint==2.3.1;python_version>='3'",
            "pylint==1.8.0;python_version<'3'",
            "pytest==4.3.1",
            "pytest-cov==2.5.1",
            "pytest-timeout==1.3.3",
            "requests==2.19.1",
            "urllib3==1.23",
            "websocket-client==0.48.0",
        ],
        'docs': [
            "pyenchant==1.6.11",
            "Sphinx==1.8.5",
            "sphinxcontrib-spelling==4.2.1",
            "sphinx-nameko-theme==0.0.3",
        ],
        'examples': [
            "nameko-sqlalchemy==0.0.1",
            "PyJWT==1.5.2",
            "moto==1.3.6",
            "bcrypt==3.1.3",
            "regex==2018.2.21"
        ]
    },
    entry_points={
        'console_scripts': [
            'nameko=nameko.cli:cli',
        ],
        'pytest11': [
            'pytest_nameko=nameko.testing.pytest'
        ]
    },
    zip_safe=True,
    license='Apache License, Version 2.0',
    classifiers=[
        "Programming Language :: Python",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Developers",
    ]
)



from __future__ import absolute_import

from contextlib import contextmanager
from logging import getLogger

from nameko.containers import get_container_cls, get_service_name
from nameko.utils.concurrency import SpawningProxy


_log = getLogger(__name__)


class ServiceRunner(object):
    """ Allows the user to serve a number of services concurrently.
    The caller can register a number of service classes with a name and
    then use the start method to serve them and the stop and kill methods
    to stop them. The wait method will block until all services have stopped.

    Example::

        runner = ServiceRunner()
        runner.add_service(Foobar)
        runner.add_service(Spam)

        add_sig_term_handler(runner.kill)

        runner.start()

        runner.wait()
    """
    def __init__(self):
        self.service_map = {}

        self.container_cls = get_container_cls()

    @property
    def service_names(self):
        return self.service_map.keys()

    @property
    def containers(self):
        return self.service_map.values()

    def add_service(self, cls):
        """ Add a service class to the runner.
        There can only be one service class for a given service name.
        Service classes must be registered before calling start()
        """
        service_name = get_service_name(cls)
        container = self.container_cls(cls)
        self.service_map[service_name] = container

    def start(self):
        """ Start all the registered services.

        A new container is created for each service using the container
        class provided in the __init__ method.

        All containers are started concurrently and the method will block
        until all have completed their startup routine.
        """
        service_names = ', '.join(self.service_names)
        _log.info('starting services: %s', service_names)

        SpawningProxy(self.containers).start()

        _log.debug('services started: %s', service_names)

    def stop(self):
        """ Stop all running containers concurrently.
        The method blocks until all containers have stopped.
        """
        service_names = ', '.join(self.service_names)
        _log.info('stopping services: %s', service_names)

        SpawningProxy(self.containers).stop()

        _log.debug('services stopped: %s', service_names)

    def kill(self):
        """ Kill all running containers concurrently.
        The method will block until all containers have stopped.
        """
        service_names = ', '.join(self.service_names)
        _log.info('killing services: %s', service_names)

        SpawningProxy(self.containers).kill()

        _log.debug('services killed: %s ', service_names)

    def wait(self):
        """ Wait for all running containers to stop.
        """
        try:
            SpawningProxy(self.containers, abort_on_error=True).wait()
        except Exception:
            # If a single container failed, stop its peers and re-raise the
            # exception
            self.stop()
            raise


@contextmanager
def run_services(*services, **kwargs):
    """ Serves a number of services for a contextual block.
    The caller can specify a number of service classes then serve them either
    stopping (default) or killing them on exiting the contextual block.


    Example::

        with run_services(Foobar, Spam) as runner:
            # interact with services and stop them on exiting the block

        # services stopped


    Additional configuration available to :class:``ServiceRunner`` instances
    can be specified through keyword arguments::

        with run_services(Foobar, Spam, kill_on_exit=True):
            # interact with services

        # services killed

    :Parameters:
        services : service definitions
            Services to be served for the contextual block
        kill_on_exit : bool (default=False)
            If ``True``, run ``kill()`` on the service containers when exiting
            the contextual block. Otherwise ``stop()`` will be called on the
            service containers on exiting the block.

    :Returns: The configured :class:`ServiceRunner` instance

    """
    kill_on_exit = kwargs.pop('kill_on_exit', False)

    runner = ServiceRunner()
    for service in services:
        runner.add_service(service)

    runner.start()

    yield runner

    if kill_on_exit:
        runner.kill()
    else:
        runner.stop()



import json

import pytest
from mock import patch
from werkzeug.wrappers import Response

from nameko.testing.utils import get_extension
from nameko.web.handlers import HttpRequestHandler, http


class ExampleService(object):
    name = "exampleservice"

    @http('GET', '/foo/<int:bar>')
    def do_get(self, request, bar):
        return 'value: {}'.format(bar)

    @http('POST', '/post')
    def do_post(self, request):
        data = json.loads(request.get_data(as_text=True))
        value = data['value']
        return value

    @http('GET,PUT,POST', '/multi')
    def do_multi(self, request):
        return request.method

    @http('GET', '/custom')
    def do_custom(self, request):
        return Response('response')

    @http('GET', '/status_code')
    def do_status_code(self, request):
        return 201, 'created'

    @http('GET', '/headers')
    def do_headers(self, request):
        return 201, {'x-foo': 'bar'}, 'created'

    @http('GET', '/fail')
    def fail(self, request):
        raise ValueError('oops')

    @http('GET', '/fail_expected', expected_exceptions=ValueError)
    def fail_expected(self, request):
        raise ValueError('oops')


class SimpleService(object):
    name = "simpleservice"

    @http('GET', '/method')
    def method(self, request):
        pass  # pragma: no cover


@pytest.fixture
def web_session(container_factory, web_config, web_session):
    container = container_factory(ExampleService)
    container.start()
    return web_session


def test_get(web_session):
    rv = web_session.get('/foo/42')
    assert rv.text == 'value: 42'

    rv = web_session.get('/foo/something')
    assert rv.status_code == 404


def test_post(web_session):
    rv = web_session.post('/post', data=json.dumps({
        'value': 'foo',
    }))
    assert rv.text == "foo"


def test_multi_get(web_session):
    rv = web_session.get('/multi')
    assert rv.text == "GET"


def test_multi_put(web_session):
    rv = web_session.put('/multi')
    assert rv.text == "PUT"


def test_multi_post(web_session):
    rv = web_session.post('/multi')
    assert rv.text == "POST"


def test_custom_response(web_session):
    rv = web_session.get('/custom')
    assert rv.text == 'response'


def test_custom_status_code(web_session):
    rv = web_session.get('/status_code')
    assert rv.text == 'created'
    assert rv.status_code == 201


def test_custom_headers(web_session):
    rv = web_session.get('/headers')
    assert rv.text == 'created'
    assert rv.status_code == 201
    assert rv.headers['x-foo'] == 'bar'


def test_broken_method(web_session):
    rv = web_session.get('/fail')
    assert rv.status_code == 500
    assert "ValueError: oops" in rv.text


def test_broken_method_expected(web_session):
    rv = web_session.get('/fail_expected')
    assert rv.status_code == 400
    assert "ValueError: oops" in rv.text


def test_bad_payload(web_session):
    rv = web_session.post('/post', data=json.dumps({'value': 23}))
    assert rv.status_code == 500
    assert "Error: TypeError: Payload must be a string. Got `23`" in rv.text


@pytest.mark.usefixtures("web_config")
def test_lifecycle(container_factory):

    container = container_factory(SimpleService)

    http = get_extension(container, HttpRequestHandler)

    with patch.object(http.server, 'register_provider') as register:
        container.start()
        register.assert_called_with(http)

    with patch.object(http.server, 'unregister_provider') as unregister:
        container.stop()
        unregister.assert_called_with(http)


class TestEntrypointArguments:

    @pytest.mark.usefixtures("web_config")
    def test_expected_exceptions_and_sensitive_arguments(
        self, container_factory
    ):

        class Boom(Exception):
            pass

        class Service(object):
            name = "service"

            @http(
                "GET", "/method",
                expected_exceptions=Boom,
                sensitive_arguments=["request"]
            )
            def method(self, request):
                pass  # pragma: no cover

        container = container_factory(Service)
        container.start()

        entrypoint = get_extension(container, HttpRequestHandler)
        assert entrypoint.expected_exceptions == Boom
        assert entrypoint.sensitive_arguments == ["request"]



from nameko.runners import ServiceRunner
from nameko.testing.utils import get_container

class ServiceA:
    name = "service_a"

class ServiceB:
    name = "service_b"

# create a runner for ServiceA and ServiceB
runner = ServiceRunner()
runner.add_service(ServiceA)
runner.add_service(ServiceB)

# ``get_container`` will return the container for a particular service
container_a = get_container(runner, ServiceA)

# start both services
runner.start()

# stop both services
runner.stop()



import warnings
from contextlib import contextmanager

from kombu import Connection
from kombu.exceptions import ChannelError
from kombu.pools import connections, producers

from nameko.constants import (
    DEFAULT_RETRY_POLICY, DEFAULT_TRANSPORT_OPTIONS, PERSISTENT
)


class UndeliverableMessage(Exception):
    """ Raised when publisher confirms are enabled and a message could not
    be routed or persisted """
    pass


@contextmanager
def get_connection(amqp_uri, ssl=None, transport_options=None):
    if not transport_options:
        transport_options = DEFAULT_TRANSPORT_OPTIONS.copy()
    conn = Connection(amqp_uri, transport_options=transport_options, ssl=ssl)

    with connections[conn].acquire(block=True) as connection:
        yield connection


@contextmanager
def get_producer(amqp_uri, confirms=True, ssl=None, transport_options=None):
    if transport_options is None:
        transport_options = DEFAULT_TRANSPORT_OPTIONS.copy()
    transport_options['confirm_publish'] = confirms
    conn = Connection(amqp_uri, transport_options=transport_options, ssl=ssl)

    with producers[conn].acquire(block=True) as producer:
        yield producer


class Publisher(object):
    """
    Utility helper for publishing messages to RabbitMQ.
    """

    use_confirms = True
    """
    Enable `confirms <http://www.rabbitmq.com/confirms.html>`_ for this
    publisher.

    The publisher will wait for an acknowledgement from the broker that
    the message was received and processed appropriately, and otherwise
    raise. Confirms have a performance penalty but guarantee that messages
    aren't lost, for example due to stale connections.
    """

    transport_options = DEFAULT_TRANSPORT_OPTIONS.copy()
    """
    A dict of additional connection arguments to pass to alternate kombu
    channel implementations. Consult the transport documentation for
    available options.
    """

    delivery_mode = PERSISTENT
    """
    Default delivery mode for messages published by this Publisher.
    """

    mandatory = False
    """
    Require `mandatory <https://www.rabbitmq.com/amqp-0-9-1-reference.html
    #basic.publish.mandatory>`_ delivery for published messages.
    """

    priority = 0
    """
    Priority value for published messages, to be used in conjunction with
    `consumer priorities <https://www.rabbitmq.com/priority.html>_`.
    """

    expiration = None
    """
    `Per-message TTL <https://www.rabbitmq.com/ttl.html>`_, in milliseconds.
    """

    serializer = "json"
    """ Name of the serializer to use when publishing messages.

    Must be registered as a
    `kombu serializer <http://bit.do/kombu_serialization>`_.
    """

    compression = None
    """ Name of the compression to use when publishing messages.

    Must be registered as a
    `kombu compression utility <http://bit.do/kombu-compression>`_.
    """

    retry = True
    """
    Enable automatic retries when publishing a message that fails due
    to a connection error.

    Retries according to :attr:`self.retry_policy`.
    """

    retry_policy = DEFAULT_RETRY_POLICY
    """
    Policy to apply when retrying message publishes, if requested.

    See :attr:`self.retry`.
    """

    declare = []
    """
    Kombu :class:`~kombu.messaging.Queue` or :class:`~kombu.messaging.Exchange`
    objects to (re)declare before publishing a message.
    """

    def __init__(
        self, amqp_uri, use_confirms=None, serializer=None, compression=None,
        delivery_mode=None, mandatory=None, priority=None, expiration=None,
        declare=None, retry=None, retry_policy=None, ssl=None, **publish_kwargs
    ):
        self.amqp_uri = amqp_uri
        self.ssl = ssl

        # publish confirms
        if use_confirms is not None:
            self.use_confirms = use_confirms

        # delivery options
        if delivery_mode is not None:
            self.delivery_mode = delivery_mode
        if mandatory is not None:
            self.mandatory = mandatory
        if priority is not None:
            self.priority = priority
        if expiration is not None:
            self.expiration = expiration

        # message options
        if serializer is not None:
            self.serializer = serializer
        if compression is not None:
            self.compression = compression

        # retry policy
        if retry is not None:
            self.retry = retry
        if retry_policy is not None:
            self.retry_policy = retry_policy

        # declarations
        if declare is not None:
            self.declare = declare

        # other publish arguments
        self.publish_kwargs = publish_kwargs

    def publish(self, payload, **kwargs):
        """ Publish a message.
        """
        publish_kwargs = self.publish_kwargs.copy()

        # merge headers from when the publisher was instantiated
        # with any provided now; "extra" headers always win
        headers = publish_kwargs.pop('headers', {}).copy()
        headers.update(kwargs.pop('headers', {}))
        headers.update(kwargs.pop('extra_headers', {}))

        use_confirms = kwargs.pop('use_confirms', self.use_confirms)
        transport_options = kwargs.pop('transport_options',
                                       self.transport_options
                                       )
        transport_options['confirm_publish'] = use_confirms

        delivery_mode = kwargs.pop('delivery_mode', self.delivery_mode)
        mandatory = kwargs.pop('mandatory', self.mandatory)
        priority = kwargs.pop('priority', self.priority)
        expiration = kwargs.pop('expiration', self.expiration)
        serializer = kwargs.pop('serializer', self.serializer)
        compression = kwargs.pop('compression', self.compression)
        retry = kwargs.pop('retry', self.retry)
        retry_policy = kwargs.pop('retry_policy', self.retry_policy)

        declare = self.declare[:]
        declare.extend(kwargs.pop('declare', ()))

        publish_kwargs.update(kwargs)  # remaining publish-time kwargs win

        with get_producer(self.amqp_uri,
                          use_confirms,
                          self.ssl,
                          transport_options,
                          ) as producer:
            try:
                producer.publish(
                    payload,
                    headers=headers,
                    delivery_mode=delivery_mode,
                    mandatory=mandatory,
                    priority=priority,
                    expiration=expiration,
                    compression=compression,
                    declare=declare,
                    retry=retry,
                    retry_policy=retry_policy,
                    serializer=serializer,
                    **publish_kwargs
                )
            except ChannelError as exc:
                if "NO_ROUTE" in str(exc):
                    raise UndeliverableMessage()
                raise

            if mandatory:
                if not use_confirms:
                    warnings.warn(
                        "Mandatory delivery was requested, but "
                        "unroutable messages cannot be detected without "
                        "publish confirms enabled."
                    )



# coding: utf-8

import pytest
from mock import Mock

from nameko.extensions import (
    DependencyProvider, Entrypoint, Extension, is_dependency, is_entrypoint,
    is_extension
)
from nameko.testing.services import entrypoint_hook
from nameko.testing.utils import get_extension


class SimpleExtension(Extension):
    pass


class SimpleDependencyProvider(DependencyProvider):
    ext = SimpleExtension()


class SimpleEntrypoint(Entrypoint):
    pass


simple = SimpleEntrypoint.decorator


class Service(object):
    name = "service"

    dep = SimpleDependencyProvider()

    @simple
    def meth1(self):
        pass  # pragma: no cover

    @simple
    def meth2(self):
        pass  # pragma: no cover


def test_entrypoint_uniqueness(container_factory):
    c1 = container_factory(Service)
    c2 = container_factory(Service)

    # entrypoint declarations are identical between containers
    c1_meth1_entrypoints = c1.service_cls.meth1.nameko_entrypoints
    c2_meth1_entrypoints = c2.service_cls.meth1.nameko_entrypoints
    assert c1_meth1_entrypoints == c2_meth1_entrypoints

    # entrypoint instances are different between containers
    c1_simple_meth1 = get_extension(c1, SimpleEntrypoint, method_name="meth1")
    c2_simple_meth1 = get_extension(c2, SimpleEntrypoint, method_name="meth1")
    assert c1_simple_meth1 != c2_simple_meth1

    # entrypoint instances are different within a container
    simple_meth1 = get_extension(c1, SimpleEntrypoint, method_name="meth1")
    simple_meth2 = get_extension(c1, SimpleEntrypoint, method_name="meth2")
    assert simple_meth1 != simple_meth2


def test_dependency_uniqueness(container_factory):
    c1 = container_factory(Service)
    c2 = container_factory(Service)

    # dependencyprovider declarations are identical between containers
    assert c1.service_cls.dep == c2.service_cls.dep

    # dependencyprovider instances are different between containers
    dep1 = get_extension(c1, SimpleDependencyProvider)
    dep2 = get_extension(c2, SimpleDependencyProvider)
    assert dep1 != dep2


def test_extension_uniqueness(container_factory):
    c1 = container_factory(Service)
    c2 = container_factory(Service)
    dep1 = get_extension(c1, SimpleDependencyProvider)
    dep2 = get_extension(c2, SimpleDependencyProvider)

    # extension declarations are identical between containers
    assert c1.service_cls.dep.ext == c2.service_cls.dep.ext

    # extension instances are different between dependencies
    assert dep1 != dep2
    assert dep1.ext != dep2.ext


def test_dependency_attr_name(container_factory):
    c1 = container_factory(Service)

    bound_dep_provider = get_extension(c1, SimpleDependencyProvider)
    assert bound_dep_provider.attr_name == 'dep'

    dep_provider_declaration = c1.service_cls.dep
    assert dep_provider_declaration.attr_name == 'dep'

    assert bound_dep_provider != dep_provider_declaration


def test_is_bound():
    container = Mock()

    ext = SimpleExtension()
    assert not ext.is_bound()
    bound = ext.bind(container)
    assert bound.is_bound()


def test_bound_extendions_cannot_be_bound():
    container = Mock()

    ext = SimpleExtension()
    bound = ext.bind(container)

    with pytest.raises(RuntimeError) as exc_info:
        bound.bind(container)
    assert str(exc_info.value) == "Cannot `bind` a bound extension."


def test_extension_defined_on_instance(container_factory):

    class ExtensionWithParams(Extension):
        def __init__(self, arg):
            self.arg = arg

    class DynamicDependencyProvider(DependencyProvider):
        def __init__(self, ext_arg):
            self.ext = ExtensionWithParams(ext_arg)

    class Service(object):
        name = "service"
        dep = DynamicDependencyProvider("argument_for_extension")

    container = container_factory(Service)
    container.start()

    assert len(container.extensions) == 2
    dyn_dep = get_extension(container, DynamicDependencyProvider)
    assert dyn_dep.ext.arg == "argument_for_extension"


def test_is_extension():
    ext = SimpleExtension()
    assert is_extension(ext)
    assert not is_extension(object)


def test_is_dependency():
    dep = SimpleDependencyProvider()
    assert is_dependency(dep)
    ext = SimpleExtension()
    assert not is_dependency(ext)


def test_is_entrypoint():
    entry = SimpleEntrypoint()
    assert is_entrypoint(entry)
    ext = SimpleExtension()
    assert not is_entrypoint(ext)


def test_entrypoint_decorator_does_not_mutate_service():

    class Service():
        name = "service"

        @simple
        def echo(self, arg):
            return arg

    service = Service()
    assert service.echo(1) == 1


@pytest.mark.parametrize("method_name, expected_args, expected_kwargs", [
    ("implicit_no_args", (), {}),
    ("explicit_no_args", (), {}),
    ("args", ("arg",), {}),
    ("kwargs", ("arg",), {"kwarg": "kwarg"}),
])
def test_entrypoint_decorator(method_name, expected_args, expected_kwargs,
                              container_factory):

    class ConfigurableEntrypoint(Entrypoint):
        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs

    configurable = ConfigurableEntrypoint.decorator

    class Service():
        name = "service"

        @configurable
        def implicit_no_args(self):
            return True

        @configurable()
        def explicit_no_args(self):
            return True

        @configurable('arg')
        def args(self):
            return True

        @configurable('arg', kwarg="kwarg")
        def kwargs(self):
            return True

    container = container_factory(Service)
    container.start()

    with entrypoint_hook(container, method_name) as method:
        assert method()

    entrypoint = get_extension(container, Entrypoint, method_name=method_name)
    assert entrypoint.args == expected_args
    assert entrypoint.kwargs == expected_kwargs


def test_extension_str():
    container = Mock()

    ext = Extension()
    assert str(ext).startswith('<Extension [unbound] at')

    bound = ext.bind(container)
    assert str(bound).startswith("<Extension at")


def test_entrypoint_str():
    container = Mock()
    container.service_name = "sérvice"

    ext = Entrypoint()
    assert str(ext).startswith('<Entrypoint [unbound] at')

    bound = ext.bind(container, "føbar")
    assert str(bound).startswith("<Entrypoint [sérvice.føbar] at")


def test_dependency_str():
    container = Mock()
    container.service_name = "sérvice"

    ext = DependencyProvider()
    assert str(ext).startswith('<DependencyProvider [unbound] at')

    bound = ext.bind(container, "føbar")
    assert str(bound).startswith("<DependencyProvider [sérvice.føbar] at")



from nameko.extensions import Entrypoint
from functools import partial

import boto3


class SqsReceive(Entrypoint):

    def __init__(self, url, region="eu-west-1", **kwargs):
        self.url = url
        self.region = region
        super(SqsReceive, self).__init__(**kwargs)

    def setup(self):
        self.client = boto3.client('sqs', region_name=self.region)

    def start(self):
        self.container.spawn_managed_thread(
            self.run, identifier="SqsReceiver.run"
        )

    def run(self):
        while True:
            response = self.client.receive_message(
                QueueUrl=self.url,
                WaitTimeSeconds=5,
            )
            messages = response.get('Messages', ())
            for message in messages:
                self.handle_message(message)

    def handle_message(self, message):
        handle_result = partial(self.handle_result, message)

        args = (message['Body'],)
        kwargs = {}

        self.container.spawn_worker(
            self, args, kwargs, handle_result=handle_result
        )

    def handle_result(self, message, worker_ctx, result, exc_info):
        # assumes boto client is thread-safe for this action, because it
        # happens inside the worker threads
        self.client.delete_message(
            QueueUrl=self.url,
            ReceiptHandle=message['ReceiptHandle']
        )
        return result, exc_info


receive = SqsReceive.decorator



AMQP_URI_CONFIG_KEY = 'AMQP_URI'
WEB_SERVER_CONFIG_KEY = 'WEB_SERVER_ADDRESS'
RPC_EXCHANGE_CONFIG_KEY = 'rpc_exchange'
SERIALIZER_CONFIG_KEY = 'serializer'
SERIALIZERS_CONFIG_KEY = 'SERIALIZERS'
ACCEPT_CONFIG_KEY = 'ACCEPT'
HEARTBEAT_CONFIG_KEY = 'HEARTBEAT'
AMQP_SSL_CONFIG_KEY = 'AMQP_SSL'
TRANSPORT_OPTIONS_CONFIG_KEY = 'TRANSPORT_OPTIONS'

MAX_WORKERS_CONFIG_KEY = 'max_workers'
PREFETCH_COUNT_CONFIG_KEY = 'PREFETCH_COUNT'
PARENT_CALLS_CONFIG_KEY = 'parent_calls_tracked'

DEFAULT_MAX_WORKERS = 10
DEFAULT_PREFETCH_COUNT = 10
DEFAULT_PARENT_CALLS_TRACKED = 10
DEFAULT_SERIALIZER = 'json'
DEFAULT_RETRY_POLICY = {'max_retries': 3}
DEFAULT_HEARTBEAT = 60
DEFAULT_TRANSPORT_OPTIONS = {
    'max_retries': 3,
    'interval_start': 2,
    'interval_step': 1,
    'interval_max': 5
}
DEFAULT_AMQP_URI = 'pyamqp://guest:guest@localhost'

CALL_ID_STACK_CONTEXT_KEY = 'call_id_stack'
AUTH_TOKEN_CONTEXT_KEY = 'auth_token'
LANGUAGE_CONTEXT_KEY = 'language'
USER_ID_CONTEXT_KEY = 'user_id'
USER_AGENT_CONTEXT_KEY = 'user_agent'

# delivery_mode
HEADER_PREFIX = "nameko"
NON_PERSISTENT = 1
PERSISTENT = 2



# coding: utf-8

import json

import pytest
import six
from mock import patch

from nameko.exceptions import (
    RemoteError, UnserializableValueError, deserialize,
    deserialize_to_instance, safe_for_serialization, serialize
)


OBJECT_REPR = repr(object)


class CustomError(Exception):
    pass


@pytest.yield_fixture
def registry():
    with patch('nameko.exceptions.registry', {}) as patched:
        yield patched


def test_serialize():

    exc = CustomError('something went wrong')

    assert serialize(exc) == {
        'exc_type': 'CustomError',
        'exc_path': 'test.test_exceptions.CustomError',
        'exc_args': ['something went wrong'],
        'value': 'something went wrong',
    }


def test_serialize_cannot_unicode():

    class CannotUnicode(object):
        def __str__(self):
            raise Exception('boom')

    bad_string = CannotUnicode()
    exc = CustomError(bad_string)

    assert serialize(exc) == {
        'exc_type': 'CustomError',
        'exc_path': 'test.test_exceptions.CustomError',
        'exc_args': ['[__unicode__ failed]'],
        'value': '[__unicode__ failed]',
    }


def test_serialize_args():
    cause = Exception('oops')
    exc = CustomError('something went wrong', cause)

    assert json.dumps(serialize(exc))


def test_deserialize_to_remote_error():

    exc = CustomError(u'something went ಠ_ಠ')
    data = serialize(exc)

    deserialized = deserialize(data)
    assert type(deserialized) == RemoteError
    assert deserialized.exc_type == "CustomError"
    assert deserialized.value == u"something went ಠ_ಠ"
    assert six.text_type(deserialized) == u"CustomError something went ಠ_ಠ"


@pytest.mark.usefixtures('registry')
def test_deserialize_to_instance():

    # register Error as deserializable to an instance
    deserialize_to_instance(CustomError)

    exc = CustomError('something went wrong')
    data = serialize(exc)

    deserialized = deserialize(data)
    assert type(deserialized) == CustomError
    assert str(deserialized) == "something went wrong"


def test_exception_name_clash():

    class MethodNotFound(Exception):
        # application exception with clashing name
        pass

    exc = MethodNotFound('application error')
    data = serialize(exc)

    deserialized = deserialize(data)
    assert type(deserialized) == RemoteError
    assert deserialized.exc_type == "MethodNotFound"
    assert deserialized.value == "application error"
    assert str(deserialized) == "MethodNotFound application error"

    from nameko.exceptions import MethodNotFound as NamekoMethodNotFound

    exc = NamekoMethodNotFound('missing')
    data = serialize(exc)

    deserialized = deserialize(data)
    assert type(deserialized) == NamekoMethodNotFound
    assert str(deserialized) == "missing"


def test_serialize_backwards_compat():

    exc = CustomError('something went wrong')
    data = serialize(exc)

    # nameko < 1.5.0 has no ``exc_path`` or ``exc_args`` keys
    del data['exc_path']
    del data['exc_args']

    deserialized = deserialize(data)
    assert type(deserialized) == RemoteError
    assert deserialized.exc_type == "CustomError"
    assert deserialized.value == "something went wrong"
    assert str(deserialized) == "CustomError something went wrong"

    # nameko < 1.1.4 has an extra ``traceback`` key
    data['traceback'] = "traceback string"

    deserialized = deserialize(data)
    assert type(deserialized) == RemoteError
    assert deserialized.exc_type == "CustomError"
    assert deserialized.value == "something went wrong"
    assert str(deserialized) == "CustomError something went wrong"


def test_unserializable_value_error():

    # normal value
    value = "value"
    exc = UnserializableValueError(value)

    assert exc.repr_value == "'value'"
    assert str(exc) == "Unserializable value: `'value'`"

    # un-repr-able value
    class CannotRepr(object):
        def __repr__(self):
            raise Exception('boom')

    bad_value = CannotRepr()
    exc = UnserializableValueError(bad_value)

    assert exc.repr_value == "[__repr__ failed]"
    assert str(exc) == "Unserializable value: `[__repr__ failed]`"


@pytest.mark.parametrize("value, safe_value", [
    ('str', 'str'),
    (object, OBJECT_REPR),
    ([object], [OBJECT_REPR]),
    ({object}, [OBJECT_REPR]),
    ({'foo': 'bar'}, {'foo': 'bar'}),
    ({object: None}, {OBJECT_REPR: 'None'}),
    ({None: object}, {'None': OBJECT_REPR}),
    ((1, 2), ['1', '2']),
    ([1, [2]], ['1', ['2']]),
])
def test_safe_for_serialization(value, safe_value):
    assert safe_for_serialization(value) == safe_value


def test_safe_for_serialization_bad_str():
    class BadStr(object):
        def __str__(self):
            raise Exception('boom')

    obj = BadStr()
    safe = safe_for_serialization(obj)
    assert isinstance(safe, six.string_types)
    assert 'failed' in safe



"""
Nameko components that can be used as standalone tools, without being hosted
inside a nameko-managed service.

Intended to be used as test utilities and external controls, for example to
initiate some action inside a nameko cluster.

Examples:


Use the RPC client to perform some addition on ``mathsservice``::

    >>> from nameko.standalone.rpc import ServiceRpcClient
    >>>
    >>> with ServiceRpcClient("mathsservice", config) as client:
    ...     result = client.add(2, 2)
    ...
    >>> print result
    4


Dispatch a ``custom_event`` as ``srcservice``::

    >>> from nameko.standalone.events import event_dispatcher
    >>>
    >>> with event_dispatcher("srcservice", config) as dispatch:
    ...     dispatch("custom_event", "msg")
    ...
    >>>

"""



import inspect
import re
from copy import deepcopy
from pydoc import locate

import six
from six.moves.urllib.parse import urlparse


REDACTED = "********"


def get_redacted_args(entrypoint, *args, **kwargs):
    """ Utility function for use with entrypoints that are marked with
    ``sensitive_arguments`` -- e.g. :class:`nameko.rpc.Rpc` and
    :class:`nameko.events.EventHandler`.

    :Parameters:
        entrypoint : :class:`~nameko.extensions.Entrypoint`
            The entrypoint that fired.
        args : tuple
            Positional arguments for the method call.
        kwargs : dict
            Keyword arguments for the method call.

    The entrypoint should have a ``sensitive_arguments`` attribute, the value
    of which is a string or tuple of strings specifying the arguments or
    partial arguments that should be redacted. To partially redact an argument,
    the following syntax is used::

        <argument-name>.<dict-key>[<list-index>]

    :Returns:
        A dictionary as returned by :func:`inspect.getcallargs`, but with
        sensitive arguments or partial arguments redacted.

    .. note::

        This function does not raise if one of the ``sensitive_arguments``
        doesn't match or partially match the calling ``args`` and ``kwargs``.
        This allows "fuzzier" pattern matching (e.g. redact a field if it is
        present, and otherwise do nothing).

        To avoid exposing sensitive arguments through a typo, it is recommend
        to test the configuration of each entrypoint with
        ``sensitive_arguments`` individually. For example:

        .. code-block:: python

            class Service(object):
                @rpc(sensitive_arguments="foo.bar")
                def method(self, foo):
                    pass

            container = ServiceContainer(Service, {})
            entrypoint = get_extension(container, Rpc, method_name="method")

            # no redaction
            foo = "arg"
            expected_foo = {'foo': "arg"}
            assert get_redacted_args(entrypoint, foo) == expected

            # 'bar' key redacted
            foo = {'bar': "secret value", 'baz': "normal value"}
            expected = {'foo': {'bar': "********", 'baz': "normal value"}}
            assert get_redacted_args(entrypoint, foo) == expected

    .. seealso::

        The tests for this utility demonstrate its full usage:
        :class:`test.test_utils.TestGetRedactedArgs`

    """
    sensitive_arguments = entrypoint.sensitive_arguments
    if isinstance(sensitive_arguments, six.string_types):
        sensitive_arguments = (sensitive_arguments,)

    method = getattr(entrypoint.container.service_cls, entrypoint.method_name)
    callargs = inspect.getcallargs(method, None, *args, **kwargs)
    del callargs['self']

    # make a deepcopy before redacting so that "partial" redacations aren't
    # applied to a referenced object
    callargs = deepcopy(callargs)

    def redact(data, keys):
        key = keys[0]
        if len(keys) == 1:
            try:
                data[key] = REDACTED
            except (KeyError, IndexError, TypeError):
                pass
        else:
            if key in data:
                redact(data[key], keys[1:])

    for variable in sensitive_arguments:
        keys = []
        for dict_key, list_index in re.findall(r"(\w+)|\[(\d+)\]", variable):
            if dict_key:
                keys.append(dict_key)
            elif list_index:
                keys.append(int(list_index))

        if keys[0] in callargs:
            redact(callargs, keys)

    return callargs


def import_from_path(path):
    """ Import and return the object at `path` if it exists.

    Raises an :exc:`ImportError` if the object is not found.
    """
    if path is None:
        return

    obj = locate(path)
    if obj is None:
        raise ImportError(
            "`{}` could not be imported".format(path)
        )

    return obj


def sanitize_url(url):
    """Redact password in urls."""
    parts = urlparse(url)
    if parts.password is None:
        return url
    host_info = parts.netloc.rsplit('@', 1)[-1]
    parts = parts._replace(netloc='{}:{}@{}'.format(
        parts.username, REDACTED, host_info))
    return parts.geturl()



from __future__ import absolute_import

import itertools
from collections import Counter

import eventlet
import pytest
from amqp.exceptions import NotFound
from eventlet.event import Event
from mock import ANY, Mock, patch
from six.moves import queue

from nameko import config
from nameko.amqp.consume import Consumer
from nameko.containers import WorkerContext
from nameko.events import (
    BROADCAST, SERVICE_POOL, SINGLETON, EventDispatcher, EventHandler,
    EventHandlerConfigurationError, event_handler
)
from nameko.exceptions import ContainerBeingKilled
from nameko.messaging import encode_to_headers
from nameko.standalone.events import event_dispatcher, get_event_exchange
from nameko.testing.services import dummy, entrypoint_hook, entrypoint_waiter
from nameko.testing.utils import DummyProvider, get_extension
from nameko.testing.waiting import wait_for_call
from nameko.utils.retry import retry


EVENTS_TIMEOUT = 5


def test_event_dispatcher(mock_container, mock_producer):

    container = mock_container
    container.service_name = "srcservice"

    service = Mock()
    worker_ctx = WorkerContext(container, service, DummyProvider("dispatch"))

    custom_retry_policy = {'max_retries': 5}

    event_dispatcher = EventDispatcher(retry_policy=custom_retry_policy).bind(
        container, attr_name="dispatch")
    event_dispatcher.setup()

    service.dispatch = event_dispatcher.get_dependency(worker_ctx)
    service.dispatch('eventtype', 'msg')

    headers = encode_to_headers(worker_ctx.context_data)

    expected_args = ('msg',)
    expected_kwargs = {
        'exchange': ANY,
        'routing_key': 'eventtype',
        'headers': headers,
        'declare': event_dispatcher.declare,
        'retry': event_dispatcher.publisher_cls.retry,
        'retry_policy': custom_retry_policy,
        'compression': event_dispatcher.publisher_cls.compression,
        'mandatory': event_dispatcher.publisher_cls.mandatory,
        'expiration': event_dispatcher.publisher_cls.expiration,
        'delivery_mode': event_dispatcher.publisher_cls.delivery_mode,
        'priority': event_dispatcher.publisher_cls.priority,
        'serializer': event_dispatcher.publisher_cls.serializer,
    }

    assert mock_producer.publish.call_count == 1
    args, kwargs = mock_producer.publish.call_args
    assert args == expected_args
    assert kwargs == expected_kwargs
    assert kwargs['exchange'].name == 'srcservice.events'


@pytest.mark.usefixtures("rabbit_config")
def test_event_handler(mock_container):

    container = mock_container
    container.service_name = "destservice"

    # test default configuration
    event_handler = EventHandler("srcservice", "eventtype").bind(container,
                                                                 "foobar")
    event_handler.setup()

    assert event_handler.queue.durable is True
    assert event_handler.queue.routing_key == "eventtype"
    assert event_handler.queue.exchange.name == "srcservice.events"

    # test service pool handler
    event_handler = EventHandler(
        "srcservice", "eventtype"
    ).bind(
        container, "foobar"
    )
    event_handler.setup()

    assert event_handler.queue.name == (
        "evt-srcservice-eventtype--destservice.foobar")
    assert event_handler.queue.exclusive is False

    # test broadcast handler with default identifier
    with patch('nameko.events.uuid') as mock_uuid:
        mock_uuid.uuid4().hex = "uuid-value"
        event_handler = EventHandler(
            "srcservice", "eventtype",
            handler_type=BROADCAST, reliable_delivery=False
        ).bind(
            container, "foobar"
        )
        event_handler.setup()

    assert event_handler.queue.name == (
        "evt-srcservice-eventtype--destservice.foobar-{}".format("uuid-value"))
    assert event_handler.queue.exclusive is True

    # test broadcast handler with custom identifier
    class BroadcastEventHandler(EventHandler):
        broadcast_identifier = "testbox"

    event_handler = BroadcastEventHandler(
        "srcservice", "eventtype", handler_type=BROADCAST
    ).bind(
        container, "foobar"
    )
    event_handler.setup()

    assert event_handler.queue.name == (
        "evt-srcservice-eventtype--destservice.foobar-{}".format("testbox"))
    assert event_handler.queue.exclusive is False

    # test singleton handler
    event_handler = EventHandler(
        "srcservice", "eventtype", handler_type=SINGLETON
    ).bind(
        container, "foobar"
    )
    event_handler.setup()

    assert event_handler.queue.name == "evt-srcservice-eventtype"
    assert event_handler.queue.exclusive is False

    # test reliable delivery
    event_handler = EventHandler(
        "srcservice", "eventtype"
    ).bind(
        container, "foobar"
    )
    event_handler.setup()

    assert event_handler.queue.auto_delete is False


@pytest.mark.usefixtures("rabbit_config")
class TestReliableDeliveryEventHandlerConfigurationError():

    def test_raises_with_default_broadcast_identity(
        self, mock_container
    ):
        container = mock_container
        container.service_name = "destservice"

        # test broadcast handler with reliable delivery
        with pytest.raises(EventHandlerConfigurationError):
            EventHandler(
                "srcservice", "eventtype",
                handler_type=BROADCAST, reliable_delivery=True
            ).bind(
                container, "foobar"
            )

    def test_no_raise_with_custom_identity(
        self, mock_container
    ):
        container = mock_container
        container.service_name = "destservice"

        # test broadcast handler with reliable delivery and custom identifier
        class BroadcastEventHandler(EventHandler):
            broadcast_identifier = "testbox"

        event_handler = BroadcastEventHandler(
            "srcservice", "eventtype",
            handler_type=BROADCAST, reliable_delivery=True
        ).bind(
            container, "foobar"
        )
        assert event_handler.reliable_delivery is True


# =============================================================================
# INTEGRATION TESTS
# =============================================================================


@pytest.fixture
def tracker():
    return Mock(events=[], workers=[])


@pytest.mark.usefixtures("rabbit_config")
def test_service_pooled_events(
    container_factory, queue_info, tracker
):

    class Base(object):

        @event_handler('srcservice', 'eventtype', handler_type=SERVICE_POOL)
        def handle(self, evt):
            tracker.track()
            tracker.events.append(evt)
            tracker.workers.append(self)

    class FooService(Base):
        name = "foo"

    class BarService(Base):
        name = "bar"

    for service_cls in (FooService, FooService, BarService):
        container = container_factory(service_cls)
        container.start()

    # foo service pool queue should have two consumers
    foo_queue_name = "evt-srcservice-eventtype--foo.handle"
    assert queue_info(foo_queue_name).consumer_count == 2

    # bar service pool queue should have one consumer
    bar_queue_name = "evt-srcservice-eventtype--bar.handle"
    assert queue_info(bar_queue_name).consumer_count == 1

    dispatch = event_dispatcher()

    count = itertools.count(start=1)
    with wait_for_call(
        tracker, 'track', callback=lambda *args: next(count) == 2
    ):
        dispatch("srcservice", "eventtype", "msg")

    # a total of two events should be received
    assert len(tracker.events) == 2

    # exactly one instance of each service should have been created
    # each should have received an event
    assert len(tracker.workers) == 2
    assert {type(worker) for worker in tracker.workers} == {
        FooService, BarService
    }


@pytest.mark.usefixtures("rabbit_config")
def test_service_pooled_events_multiple_handlers(
    container_factory, queue_info, tracker
):

    class Service(object):
        name = "double"

        def handle(self, evt):
            tracker.events.append(evt)
            tracker.workers.append(self)

        @event_handler('srcservice', 'eventtype', handler_type=SERVICE_POOL)
        def handle_1(self, evt):
            self.handle(evt)

        @event_handler('srcservice', 'eventtype', handler_type=SERVICE_POOL)
        def handle_2(self, evt):
            self.handle(evt)

    container = container_factory(Service)
    container.start()

    # we should have two queues with a consumer each
    queue_1_name = "evt-srcservice-eventtype--double.handle_1"
    assert queue_info(queue_1_name).consumer_count == 1
    queue_2_name = "evt-srcservice-eventtype--double.handle_2"
    assert queue_info(queue_2_name).consumer_count == 1

    dispatch = event_dispatcher()

    with entrypoint_waiter(container, 'handle_1'):
        with entrypoint_waiter(container, 'handle_2'):
            dispatch("srcservice", "eventtype", "msg")

    # each handler should have received the event
    assert len(tracker.events) == 2

    # two worker instances would have been created to deal with the handling
    assert len(tracker.workers) == 2
    assert {type(worker) for worker in tracker.workers} == {Service}


@pytest.mark.usefixtures("rabbit_config")
def test_singleton_events(container_factory, queue_info, tracker):

    class Base(object):

        @event_handler('srcservice', 'eventtype', handler_type=SINGLETON)
        def handle(self, evt):
            tracker.track()
            tracker.events.append(evt)
            tracker.workers.append(self)

    class FooService(Base):
        name = "foo"

    class BarService(Base):
        name = "bar"

    for service_cls in (FooService, FooService, BarService):
        container = container_factory(service_cls)
        container.start()

    # the singleton queue should have three consumers
    assert queue_info("evt-srcservice-eventtype").consumer_count == 3

    dispatch = event_dispatcher()

    count = itertools.count(start=1)
    with wait_for_call(
        tracker, 'track', callback=lambda *args: next(count) == 1
    ):
        dispatch("srcservice", "eventtype", "msg")

    # exactly one event should have been received
    assert len(tracker.events) == 1

    # one lucky handler should have received the event
    assert len(tracker.workers) == 1
    assert {type(worker) for worker in tracker.workers}.issubset(
        {FooService, BarService}
    )


@pytest.mark.usefixtures("rabbit_config")
def test_broadcast_events(
    container_factory, get_vhost, queue_info, tracker, rabbit_manager
):

    vhost = get_vhost(config['AMQP_URI'])

    class Base(object):

        @event_handler(
            'srcservice', 'eventtype',
            handler_type=BROADCAST, reliable_delivery=False
        )
        def handle(self, evt):
            tracker.track()
            tracker.events.append(evt)
            tracker.workers.append(self)

    class FooService(Base):
        name = "foo"

    class BarService(Base):
        name = "bar"

    for service_cls in (FooService, FooService, BarService):
        container = container_factory(service_cls)
        container.start()

    # each broadcast queue should have one consumer
    queues = rabbit_manager.get_queues(vhost)
    queue_names = [queue['name'] for queue in queues
                   if queue['name'].startswith("evt-srcservice-eventtype-")]

    assert len(queue_names) == 3
    for name in queue_names:
        # TODO can use queue_info here once
        # https://github.com/nameko/nameko/pull/484 lands and we drop the
        # exclusive flag on broadcast handlers
        # assert queue_info(name).consumer_count == 1
        queue = rabbit_manager.get_queue(vhost, name)
        assert len(queue['consumer_details']) == 1

    dispatch = event_dispatcher()

    count = itertools.count(start=1)
    with wait_for_call(
        tracker, 'track', callback=lambda *args: next(count) == 3
    ):
        dispatch("srcservice", "eventtype", "msg")

    # a total of three events should be received
    assert len(tracker.events) == 3

    # all three handlers should receive the event, but they're only of two
    # different types
    assert len(tracker.workers) == 3
    worker_counts = Counter([type(worker) for worker in tracker.workers])
    assert worker_counts[FooService] == 2
    assert worker_counts[BarService] == 1


@pytest.mark.usefixtures("rabbit_config")
def test_requeue_on_error(container_factory, queue_info, tracker):

    class Service(object):
        name = "requeue"

        @event_handler('srcservice', 'eventtype', requeue_on_error=True)
        def handle(self, evt):
            tracker.track()
            tracker.events.append(evt)
            tracker.workers.append(self)
            raise Exception("Error")

    container = container_factory(Service)
    container.start()

    # the queue should been created and have one consumer
    queue_name = "evt-srcservice-eventtype--requeue.handle"
    assert queue_info(queue_name).consumer_count == 1

    dispatch = event_dispatcher()

    counter = itertools.count(start=1)
    with entrypoint_waiter(
        container, 'handle', callback=lambda *args: next(counter) > 1
    ):
        dispatch("srcservice", "eventtype", "msg")

    # the event will be received multiple times as it gets requeued and then
    # consumed again
    assert len(tracker.events) > 1

    # multiple instances of the service should have been instantiated
    assert len(tracker.workers) > 1


@pytest.mark.usefixtures("rabbit_config")
def test_reliable_delivery(container_factory, queue_info, tracker):
    """ Events sent to queues declared by ``reliable_delivery`` handlers
    should be received even if no service was listening when they were
    dispatched.
    """
    class Service(object):
        name = "reliable"

        @event_handler('srcservice', 'eventtype', reliable_delivery=True)
        def handle(self, evt):
            tracker.track()
            tracker.events.append(evt)
            tracker.workers.append(self)

    container = container_factory(Service)
    container.start()

    # test queue created, with one consumer
    queue_name = "evt-srcservice-eventtype--reliable.handle"
    assert queue_info(queue_name).consumer_count == 1

    dispatch = event_dispatcher()

    # dispatch an event
    with entrypoint_waiter(container, 'handle'):
        dispatch("srcservice", "eventtype", "msg_1")

    assert tracker.events == ["msg_1"]

    # stop container, check queue still exists, without consumers
    container.stop()

    @retry
    def consumer_removed():
        assert queue_info(queue_name).consumer_count == 0
    consumer_removed()

    # dispatch another event while nobody is listening
    dispatch("srcservice", "eventtype", "msg_2")

    # verify the message gets queued
    assert queue_info(queue_name).message_count == 1

    # start another container
    container = container_factory(Service)
    with entrypoint_waiter(container, 'handle'):
        container.start()

    # check the new service to collects the pending event
    assert tracker.events == ["msg_1", "msg_2"]


@pytest.mark.usefixtures("rabbit_config")
def test_unreliable_delivery(container_factory, queue_info, tracker):
    """ Events sent to queues declared by non- ``reliable_delivery`` handlers
    should be lost if no service was listening when they were dispatched.
    """

    class UnreliableService(object):
        name = "unreliable"

        @event_handler('srcservice', 'eventtype', reliable_delivery=False)
        def handle(self, evt):
            tracker.track()
            tracker.events.append(evt)
            tracker.workers.append(self)

    unreliable_container = container_factory(UnreliableService)
    unreliable_container.start()

    class ReliableService(object):
        name = "reliable"

        @event_handler('srcservice', 'eventtype', reliable_delivery=True)
        def handle(self, evt):
            tracker.track()
            tracker.events.append(evt)
            tracker.workers.append(self)

    reliable_container = container_factory(ReliableService)
    reliable_container.start()

    # test unreliable queue created, with one consumer
    queue_name = "evt-srcservice-eventtype--unreliable.handle"
    assert queue_info(queue_name).consumer_count == 1

    dispatch = event_dispatcher()

    # dispatch an event
    count = itertools.count(start=1)
    with wait_for_call(
        tracker, 'track', callback=lambda *args: next(count) == 2
    ):
        dispatch("srcservice", "eventtype", "msg_1")

    assert tracker.events == ["msg_1", "msg_1"]

    # test that both services received it
    assert len(tracker.workers) == 2
    worker_counts = Counter([type(worker) for worker in tracker.workers])
    assert worker_counts[ReliableService] == 1
    assert worker_counts[UnreliableService] == 1

    # stop container, test queue deleted
    unreliable_container.stop()
    with pytest.raises(NotFound):
        # queue_info may not raise on the first call
        for _ in range(3):
            queue_info(queue_name)
            eventlet.sleep(1)  # pragma: no cover

    # dispatch a second event while nobody is listening
    count = itertools.count(start=1)
    with wait_for_call(
        tracker, 'track', callback=lambda *args: next(count) == 1
    ):
        dispatch("srcservice", "eventtype", "msg_2")

    # start another container
    unreliable_container = container_factory(UnreliableService)
    unreliable_container.start()

    # verify the queue is recreated, with one consumer
    assert queue_info(queue_name).consumer_count == 1

    # dispatch a third event
    count = itertools.count(start=1)
    with wait_for_call(
        tracker, 'track', callback=lambda *args: next(count) == 2
    ):
        dispatch("srcservice", "eventtype", "msg_3")

    # verify that the "unreliable" handler didn't receive the message sent
    # when there wasn't an instance running
    assert tracker.events == ["msg_1", "msg_1", "msg_2", "msg_3", "msg_3"]
    worker_counts = Counter([type(worker) for worker in tracker.workers])
    assert worker_counts[ReliableService] == 3
    assert worker_counts[UnreliableService] == 2


@pytest.mark.usefixtures("rabbit_config")
def test_dispatch_to_rabbit(rabbit_manager, get_vhost, mock_container):

    vhost = get_vhost(config['AMQP_URI'])

    container = mock_container
    container.shared_extensions = {}
    container.service_name = "srcservice"

    service = Mock()
    worker_ctx = WorkerContext(container, service, DummyProvider())

    dispatcher = EventDispatcher().bind(container, 'dispatch')
    dispatcher.setup()
    dispatcher.start()

    # we should have an exchange but no queues
    exchanges = rabbit_manager.get_exchanges(vhost)
    queues = rabbit_manager.get_queues(vhost)
    assert "srcservice.events" in [exchange['name'] for exchange in exchanges]
    assert queues == []

    # manually add a queue to capture the events
    rabbit_manager.create_queue(vhost, "event-sink", auto_delete=True)
    rabbit_manager.create_queue_binding(
        vhost,
        "srcservice.events",
        "event-sink",
        routing_key="eventtype"
    )

    service.dispatch = dispatcher.get_dependency(worker_ctx)
    service.dispatch("eventtype", "msg")

    # test event received on manually added queue
    messages = rabbit_manager.get_messages(vhost, "event-sink")
    assert ['"msg"'] == [msg['payload'] for msg in messages]


class TestHandlerConfigurability(object):
    """
    Test and demonstrate configuration options for the EventHandler
    """

    @pytest.mark.usefixtures("memory_rabbit_config")
    def test_heartbeat(self, mock_container):
        mock_container.service_name = "service"

        value = 999

        handler = EventHandler(
            "service", "event", heartbeat=value
        ).bind(mock_container, "method")
        handler.setup()

        assert handler.consumer.connection.heartbeat == value

    @pytest.mark.usefixtures("memory_rabbit_config")
    def test_prefetch_count(self, mock_container):
        mock_container.service_name = "service"

        value = 999

        handler = EventHandler(
            "service", "event", prefetch_count=value
        ).bind(mock_container, "method")
        handler.setup()

        assert handler.consumer.prefetch_count == value

    @pytest.mark.usefixtures("memory_rabbit_config")
    def test_accept(self, mock_container):
        mock_container.service_name = "service"

        value = ['yaml', 'json']

        handler = EventHandler(
            "service", "event", accept=value
        ).bind(mock_container, "method")
        handler.setup()

        assert handler.consumer.accept == value


class TestDispatcherConfigurability(object):
    """
    Test and demonstrate configuration options for the EventDispatcher
    """

    @pytest.yield_fixture
    def get_producer(self):
        with patch('nameko.amqp.publish.get_producer') as get_producer:
            yield get_producer

    @pytest.fixture
    def producer(self, get_producer):
        producer = get_producer().__enter__.return_value
        # make sure we don't raise UndeliverableMessage if mandatory is True
        producer.channel.returned_messages.get_nowait.side_effect = queue.Empty
        return producer

    @pytest.mark.usefixtures("memory_rabbit_config")
    @pytest.mark.parametrize("parameter", [
        # delivery options
        'delivery_mode', 'mandatory', 'priority', 'expiration',
        # message options
        'serializer', 'compression',
        # retry policy
        'retry', 'retry_policy',
        # other arbitrary publish kwargs
        'correlation_id', 'user_id', 'bogus_param'
    ])
    def test_regular_parameters(
        self, parameter, mock_container, producer
    ):
        """ Verify that most parameters can be specified at instantiation time.
        """
        mock_container.service_name = "service"

        worker_ctx = Mock()
        worker_ctx.context_data = {}

        value = Mock()

        dispatcher = EventDispatcher(
            **{parameter: value}
        ).bind(mock_container, "dispatch")
        dispatcher.setup()

        dispatch = dispatcher.get_dependency(worker_ctx)

        dispatch("event-type", "event-data")
        assert producer.publish.call_args[1][parameter] == value

    @pytest.mark.usefixtures("memory_rabbit_config")
    @pytest.mark.usefixtures('predictable_call_ids')
    def test_headers(self, mock_container, producer):
        """ Headers can be provided at instantiation time, and are merged with
        Nameko headers.
        """
        mock_container.service_name = "service"

        # use a real worker context so nameko headers are generated
        service = Mock()
        entrypoint = Mock(method_name="method")
        context_data = {'foo': 'bar'}
        worker_ctx = WorkerContext(
            mock_container, service, entrypoint, data=context_data
        )

        value = {'foo': 'bar'}

        dispatcher = EventDispatcher(
            **{'headers': value}
        ).bind(mock_container, "dispatch")
        dispatcher.setup()

        dispatch = dispatcher.get_dependency(worker_ctx)

        # context data changes are reflected up to the point of dispatching
        worker_ctx.data['foo'] = 'changed-bar'

        expected_headers = {
            'nameko.foo': 'changed-bar',
            'nameko.call_id_stack': ['service.method.0'],
            'foo': 'bar',
        }

        dispatch("event-type", "event-data")

        _, kwargs = producer.publish.call_args
        assert kwargs['headers'] == expected_headers

    @pytest.mark.usefixtures("memory_rabbit_config")
    def test_restricted_parameters(
        self, mock_container, producer
    ):
        """ Verify that providing routing parameters at instantiation
        time has no effect.
        """
        mock_container.service_name = "service"

        worker_ctx = Mock()
        worker_ctx.context_data = {}

        exchange = Mock()
        routing_key = Mock()

        dispatcher = EventDispatcher(
            exchange=exchange,
            routing_key=routing_key,
        ).bind(mock_container, "dispatch")
        dispatcher.setup()

        dispatch = dispatcher.get_dependency(worker_ctx)

        event_exchange = get_event_exchange("service")
        event_type = "event-type"

        dispatch(event_type, "event-data")

        assert producer.publish.call_args[1]['exchange'] == event_exchange
        assert producer.publish.call_args[1]['routing_key'] == event_type


class TestContainerBeingKilled(object):

    @pytest.fixture
    def dispatch(self, rabbit_config):
        return event_dispatcher()

    def test_container_killed(self, container_factory, dispatch):
        class Service(object):
            name = "service"

            @event_handler("service", "eventtype")
            def method(self, event_data):
                pass  # pragma: no cover

        container = container_factory(Service)
        container.start()

        # check message is requeued if container throws ContainerBeingKilled
        with patch.object(container, 'spawn_worker') as spawn_worker:
            spawn_worker.side_effect = ContainerBeingKilled()

            with wait_for_call(Consumer, 'requeue_message'):
                dispatch("service", "eventtype", "payload")


class TestSSL(object):

    @pytest.fixture(params=[True, False])
    def rabbit_ssl_options(self, request, rabbit_ssl_options):
        verify_certs = request.param
        if verify_certs is False:
            # remove certificate paths from config
            options = True
        else:
            options = rabbit_ssl_options
        return options

    @pytest.mark.usefixtures("rabbit_ssl_config")
    def test_event_handler_over_ssl(self, container_factory, rabbit_uri):
        class Service(object):
            name = "service"

            @event_handler("service", "event")
            def echo(self, event_data):
                return event_data

        container = container_factory(Service)
        container.start()

        dispatch = event_dispatcher(uri=rabbit_uri, ssl=None)

        with entrypoint_waiter(container, 'echo') as result:
            dispatch("service", "event", "payload")
        assert result.get() == "payload"

    @pytest.mark.usefixtures("rabbit_config")
    def test_event_dispatcher_over_ssl(
        self, container_factory, rabbit_ssl_uri, rabbit_ssl_options
    ):
        class Dispatcher(object):
            name = "dispatch"

            dispatch = EventDispatcher(uri=rabbit_ssl_uri, ssl=rabbit_ssl_options)

            @dummy
            def method(self, payload):
                return self.dispatch("event-type", payload)

        class Handler(object):
            name = "handler"

            @event_handler("dispatch", "event-type")
            def echo(self, payload):
                return payload

        dispatcher = container_factory(Dispatcher)
        dispatcher.start()

        handler = container_factory(Handler)
        handler.start()

        with entrypoint_waiter(handler, 'echo') as result:
            with entrypoint_hook(dispatcher, 'method') as dispatch:
                dispatch("payload")
        assert result.get() == "payload"


@pytest.mark.usefixtures("rabbit_config")
def test_stop_with_active_worker(container_factory, queue_info):
    """ Test behaviour when we stop a container with an active worker.

    Expect the consumer to stop and the message be requeued, but the container
    to continue running the active worker until it completes.

    This is not desirable behaviour but it is consistent with the old
    implementation. It would be better to stop the consumer but keep the
    channel alive until the worker has completed and the message can be
    ack'd, but we can't do that with kombu or without per-entrypoint worker
    pools.
    """

    block = Event()

    class Service(object):
        name = "service"

        @event_handler("service", "event")
        def method(self, event_data):
            block.wait()

    container = container_factory(Service)
    container.start()

    dispatch = event_dispatcher()
    dispatch("service", "event", "payload")

    gt = eventlet.spawn(container.stop)

    @retry
    def consumer_removed():
        info = queue_info('evt-service-event--service.method')
        assert info.consumer_count == 0
        assert info.message_count == 1

    consumer_removed()

    assert not gt.dead
    eventlet.sleep(10)
    block.send(True)

    gt.wait()
    assert gt.dead


class TestEntrypointArguments:

    @pytest.mark.usefixtures("rabbit_config")
    def test_expected_exceptions_and_sensitive_arguments(
        self, container_factory
    ):

        class Boom(Exception):
            pass

        class Service(object):
            name = "service"

            @event_handler(
                "service", "event",
                expected_exceptions=Boom, sensitive_arguments=["event_data"]
            )
            def method(self, event_data):
                pass  # pragma: no cover

        container = container_factory(Service)
        container.start()

        entrypoint = get_extension(container, EventHandler)
        assert entrypoint.expected_exceptions == Boom
        assert entrypoint.sensitive_arguments == ["event_data"]
